from renpy.display.matrix cimport Matrix

cdef class Polygon:

    # The number of points in the polygon.
    cdef int points

    # The stride of the polygon - the number of floats comprising a data
    # point.
    cdef int stride

    # The data in a polygon.
    cdef float *data

    cpdef Polygon copy(Polygon self)

    cpdef void offset_inplace(Polygon self, double x, double y, double z)
    cpdef Polygon offset(Polygon self, double x, double y, double z)

    cpdef void multiply_matrix_inplace(Polygon self, Matrix matrix)
    cpdef Polygon multiply_matrix(Polygon self, Matrix matrix)

    cpdef void perspective_divide_inplace(Polygon self)
    cpdef Polygon perspective_divide(Polygon self)

    cpdef Polygon intersect(Polygon self, Polygon other)

# Constructors for particularly useful polygons.
cpdef Polygon rectangle(double l, double t, double r, double b)
cpdef Polygon texture_rectangle(double pl, double pt, double pr, double pb, double tl, double tt, double tr, double tb)

cdef class Mesh:

    # The total number of points.
    cdef public int points

    # The stride - the amount of data per point, in floats.
    cdef public int stride

    # A map from an attribute to the offset of that attribute.
    cdef public dict attributes

    # A list of polygons that comprise the mesh.
    cdef public list polygons

    # An array that contains all the data from the points. This is NULL
    # until it's created.
    cdef float *data

    cdef float *get_data(self, int offset)

    # The count of polygons that comprise the mesh. This is  only valid after
    # get_data is called.
    cdef public int polygon_count

    # If all the polygons in the mesh have the same number of points, this is
    # that number of points. Else, it's 0.
    cdef public int polygon_points

    cpdef Mesh copy(Mesh self)

    cpdef void offset_inplace(Mesh self, double x, double y, double z)
    cpdef Mesh offset(Mesh self, double x, double y, double z)

    cpdef void multiply_matrix_inplace(Mesh self, Matrix matrix)
    cpdef Mesh multiply_matrix(Mesh self, Matrix matrix)

    cpdef void perspective_divide_inplace(Mesh self)
    cpdef Mesh perspective_divide(Mesh self)

    cpdef Mesh intersect(Mesh self, Mesh other)
    cpdef Mesh crop(Mesh self, Polygon other)




#@PydevCodeAnalysisIgnore
#cython: profile=False
# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function

DEF ANGLE = False


from sdl2 cimport *
from pygame_sdl2 cimport *
import_pygame_sdl2()

from libc.stdlib cimport malloc, free
from libc.string cimport memcpy

import sys
import time
import collections
import weakref
import math

import renpy
from renpy.gl2.uguugl cimport *
from renpy.gl2.gl2draw cimport GL2Draw
from renpy.gl2.gl2geometry cimport rectangle, texture_rectangle, Mesh

from renpy.display.matrix cimport Matrix

# This has different names in GL and GLES, but the same value.
cdef GLenum RGBA8 = 0x8058

################################################################################

cdef class TextureLoader:

    def __init__(TextureLoader self, GL2Draw draw):
        self.allocated = set()
        self.free_list = [ ]
        self.total_texture_size = 0
        self.texture_load_queue = weakref.WeakSet()
        self.draw = draw

        self.ftl_program = draw.shader_cache.get(("renpy.ftl",))

    def quit(self):
        """
        Gets rid of this TextureLoader.
        """

        cdef GLuint texnums[1]

        for texture_number in self.allocated:
            texnums[0] = texture_number
            glDeleteTextures(1, texnums)

    def get_texture_size(self):
        """
        Returns the amount of memory locked up in textures.
        """

        return self.total_texture_size, len(self.allocated)

    def load_one_surface(self, surf, bl, bt, br, bb):
        """
        Converts a surface into a texture.
        """

        size = surf.get_size()

        rv = Texture(size, self)
        rv.from_surface(surf)

        if bl or bt or br or bb:
            w, h = size

            mesh = Mesh()
            mesh.add_attribute("aTexCoord", 2)

            pw = w - bl - br
            ph = h - bt - bb

            if (w and h):

                mesh.add_texture_rectangle(
                    0.0, 0.0, pw, ph,
                    1.0 * bl / w, 1.0 * bt / h, 1.0 - 1.0 * br / w, 1.0 - 1.0 * bb / h)

            rv = Model((pw, ph), mesh, ("renpy.texture",), { "uTex0" : rv })

        return rv

    def texture_axis(self, length, limit, border):
        """
        Splits `length` up into multiple textures.

        This returns a series of (offset, width, left/top border, right/bottom border) tuples.
        """

        if length <= limit:
            return [ (0, length, 0, 0) ]

        elif length <= 2 * (limit - border):

            right = length // 2
            left = length - right

            return [ (0, left, 0, border), (left, right, border, 0) ]

        else:
            tiles = math.ceil(1.0 * length / (limit - 2 * border))
            tile_length = length / tiles
            tiles = int(tiles)

            rv = [ ]

            for i in xrange(tiles):
                start = int(i * tile_length)
                end = int((i + 1) * tile_length)

                if i > 0:
                    left = border
                else:
                    left = 0

                if i < tiles - 1:
                    right = border
                else:
                    right = 0

                rv.append((start, end - start, left, right))

        return rv


    def load_surface(self, surf):
        border = 1

        size = surf.get_size()
        w, h = size

        if (w <= self.max_texture_width) and (h <= self.max_texture_height):
            return self.load_one_surface(surf, 0, 0, 0, 0)

        htiles = self.texture_axis(w, self.max_texture_width, border)
        vtiles = self.texture_axis(h, self.max_texture_height, border)

        rv = renpy.display.render.Render(w, h)

        for ty, th, bt, bb in vtiles:
            for tx, tw, bl, br in htiles:
                ss = surf.subsurface((tx - bl, ty - bt, tw + bl + br, th + bt + bb))
                t = self.load_one_surface(ss, bl, bt, br, bb)
                rv.blit(t, (tx, ty))

        return rv

    def render_to_texture(self, what):
        """
        Renders `what` to a texture.
        """

        rv = Texture(what.get_size(), self)
        rv.from_render(what)
        return rv


    def cleanup(self):
        """
        This is called once per frame, to free textures that are no longer used.
        """

        cdef GLuint texnums[1]

        for texture_number in self.free_list:
            texnums[0] = texture_number
            glDeleteTextures(1, texnums)
            self.allocated.remove(texture_number)

        self.free_list = [ ]


    def ready_one_texture(self):
        """
        Called by GL2Draw to implement ready_one_texture.
        """

        while True:

            try:
                tex = self.texture_load_queue.pop()
            except KeyError:
                return False

            if not tex.loaded:
                tex.load()
                return True

        return False

cdef class GLTexture:
    """
    This class represents an OpenGL texture that needs to be loaded by
    Ren'Py. It's responsible for handling deferred loading of textures,
    and using the Python reference counting system to free textures that
    are no longer required.
    """

    def __init__(GLTexture self, size, TextureLoader loader):

        cdef unsigned char *pixels
        cdef unsigned char *data
        cdef unsigned char *p

        # The width and height of this texture.
        self.width, self.height = size

        # The number of the OpenGL texture this texture object
        # represents.
        self.number = 0

        # True if the texture has been loaded into OpenGL, False otherwise.
        self.loaded = False

        # Used for loading surfaces.
        self.data = NULL
        self.surface = None

        # Update the loader.
        self.loader = loader
        self.loader.total_texture_size += self.width * self.height * 4

    def from_surface(GLTexture self, surface):
        """
        Called to indicate this texture should be loaded from a surface.
        """

        # Make a copy of the texture data.
        cdef SDL_Surface *s
        s = PySurface_AsSurface(surface)

        pitch_pixels = s.pitch / 4
        margin = pitch_pixels - s.w

        if s.w and s.h and (margin < 64) and s.w < self.loader.max_texture_width:
            # In-place path.

            self.data = NULL
            self.surface = surface

        else:
            # Copying path.

            pixels = <unsigned char *> s.pixels
            data = <unsigned char *> malloc(s.h * s.w * 4)
            p = data

            if s.pitch == s.w * 4:
                memcpy(p, pixels, s.h * s.w * 4)
            else:
                for 0 <= i < s.h:
                    memcpy(p, pixels, s.w * 4)

                    pixels += s.pitch
                    p += (s.w * 4)

            self.data = data
            self.surface = None

        self.loader.texture_load_queue.add(self)

    def from_render(GLTexture self, what):
        """
        This renders `what` to this texture.
        """

        cw, ch = size = what.get_size()

        loader = self.loader
        draw = self.loader.draw

        # The visible size of the texture.
        tw = min(int(math.ceil(cw * draw.draw_per_virt)), loader.max_texture_width)
        th = min(int(math.ceil(ch * draw.draw_per_virt)), loader.max_texture_height)

        cdef GLuint premultiplied

        glGenTextures(1, &premultiplied)

        # Bind the framebuffer.
        draw.change_fbo(draw.fbo)

        # Set up the viewport.
        glViewport(0, 0, tw, th)

        # Clear the screen.
        clear_r, clear_g, clear_b = renpy.color.Color(renpy.config.gl_clear_color).rgb
        glClearColor(clear_r, clear_g, clear_b, 0.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # Project the child from virtual space to the screen space.
        cdef Matrix transform
        transform = renpy.display.matrix.texture_projection(cw, ch)
        if what.reverse:
            transform *= what.reverse

        # Set up the default modes.
        glEnable(GL_BLEND)
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)

        context = renpy.gl2.gl2draw.GL2DrawingContext(draw)
        context.draw(what, transform)

        glBindTexture(GL_TEXTURE_2D, premultiplied)
        glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, tw, th, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)

        self.number = premultiplied
        self.loader.allocated.add(self.number)

        self.loaded = True

    def __repr__(self):
        return "<GLTexture {}x{} {}>".format(self.width, self.height, self.number)

    def load_gltexture(GLTexture self):
        """
        Loads this texture. When it's loaded, generation and number are set,
        and the texture is ready to use.
        """

        cdef GLuint tex
        cdef GLuint premultiplied
        cdef Program program
        cdef SDL_Surface *s

        if self.loaded:
            return

        draw = self.loader.draw

        # Generate the old textures.
        glGenTextures(1, &tex)
        glGenTextures(1, &premultiplied)

        # Bind the framebuffer.
        draw.change_fbo(draw.fbo)

        # Load the pixel data into tex, and set it up for drawing.
        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, tex)

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)

        mesh = Mesh()
        mesh.add_attribute("aTexCoord", 2)

        # Load the texture through zero-copy and normal paths.
        if self.surface is not None:
            s = PySurface_AsSurface(self.surface)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, s.pitch / 4, self.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, s.pixels)
            mesh.add_texture_rectangle(-1.0, -1.0, 1.0, 1.0, 0.0, 0.0, 4.0 * s.w / s.pitch, 1.0)
        else:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.width, self.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, self.data)
            mesh.add_texture_rectangle(-1.0, -1.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0)

        # Set up the viewport.
        glViewport(0, 0, self.width, self.height)

        # Set up the blend mode for premultiplication.
        glEnable(GL_BLEND)
        glBlendFuncSeparate(GL_SRC_ALPHA, GL_ZERO, GL_ONE, GL_ZERO)

        # Draw.
        program = self.loader.ftl_program
        program.start()
        program.set_uniform("uTex0", tex)
        program.draw(mesh)
        program.finish()

        # Create premultiplied.
        glBindTexture(GL_TEXTURE_2D, premultiplied)
        glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, self.width, self.height, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)

        # Delete tex.
        glDeleteTextures(1, &tex)

        # Store the loaded texture.
        self.number = premultiplied
        self.loader.allocated.add(self.number)

        self.loaded = True

        # Free the data memory.
        if self.data != NULL:
            free(self.data)
            self.data = NULL

        self.surface = None

    def __dealloc__(self):

        if self.data:
            free(self.data)
            self.data = NULL

    def __del__(self):
        try:
            if self.loaded:
                self.loader.free_list.append(self.number)

            self.loader.total_texture_size -= self.width * self.height * 4
        except TypeError:
            pass # Let's not error on shutdown.


################################################################################

class Model(object):
    """
    A Model can be placed in the render tree, and contains the information
    required to be drawn to a screen.
    """

    def __init__(self, size, mesh, shaders, uniforms):
        # The size of this model's bounding box, in virtual pixels.
        self.size = size

        # The mesh.
        self.mesh = mesh

        # A tuple giving the shaders used with this model.
        self.shaders = shaders

        # Either a dictionary giving uniforms associated with this model,
        # or None.
        self.uniforms = uniforms

        # The cached_texture that comes from this model.
        self.cached_texture = None

    def load(self):
        """
        Loads the textures associated with this model.
        """

        for i in self.uniforms.itervalues():
            if isinstance(i, GLTexture):
                i.load_gltexture()

    def program_uniforms(self, shader):
        """
        Called by the rest of the drawing code to set up the textures associated
        with this model.
        """

        shader.set_uniforms(self.uniforms)

    def get_size(self):
        """
        Returns the size of this Model.
        """

        return self.size

    def subsurface(self, rect):
        """
        Creates Model that fits within `rect`.
        """

        x, y, w, h = rect

        mesh = self.mesh.crop(rectangle(x, y, x+w, y+h))

        if mesh is self.mesh:
            if (x == 0) and (y == 0):
                return self

            mesh = mesh.offset(-x, -y, 0)
        else:
            mesh.offset_inplace(-x, -y, 0)

        rv = Model((w, h), self.mesh, self.shaders, self.uniforms)
        rv.mesh = mesh

        return rv


class Texture(GLTexture, Model):
    """
    A texture is Model and a GLTexture at the same time.

    This also implies that the Model has texture coordinates that range from
    (0.0, 0.0) to (1.0, 1.0) - and hence, that the GLTexture can be used to
    represent it.
    """

    def __init__(self, size, loader):

        GLTexture.__init__(self, size, loader)

        mesh = Mesh()
        mesh.add_attribute("aTexCoord", 2)

        w, h = size

        mesh.add_texture_rectangle(
            0.0, 0.0, w, h,
            0.0, 0.0, 1.0, 1.0,
            )

        Model.__init__(self, size, mesh, ("renpy.texture",), None)

    def load(self):
        self.load_gltexture()

    def program_uniforms(self, shader):
        shader.set_uniform("uTex0", self)

    def subsurface(self, rect):
        rv = Model.subsurface(self, rect)
        if rv is not self:
            rv.uniforms = { "uTex0" : self }
        return rv



from __future__ import print_function

from libc.string cimport memset
from libc.math cimport sin, cos, M_PI as pi

cdef float *aligned_1 = [ 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 ]
cdef float *aligned_2 = [ 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1 ]

fields = [
    "xdx", "xdy", "xdz", "xdw",
    "ydx", "ydy", "ydz", "ydw",
    "zdx", "zdy", "zdz", "zdw",
    "wdx", "wdy", "wdz", "wdw",
    ]


cdef class Matrix:
    """
    Represents a 4x4 matrix.
    """

    def __init__(Matrix self, l):

        memset(self.m, 0, sizeof(float) * 16)

        if l is None:
            return

        cdef int lenl = len(l)

        if lenl == 4:
            (self.xdx, self.xdy,
             self.ydx, self.ydy) = l
            self.ydy = 1.0
            self.wdw = 1.0

        elif lenl == 9:
            (self.xdx, self.xdy, self.xdz,
             self.ydx, self.ydy, self.ydz,
             self.zdx, self.zdy, self.zdz) = l
            self.wdw = 1.0

        elif lenl == 16:
            (self.xdx, self.xdy, self.xdz, self.xdw,
             self.ydx, self.ydy, self.ydz, self.ydw,
             self.zdx, self.zdy, self.zdz, self.zdw,
             self.wdx, self.wdy, self.wdz, self.wdw) = l

        else:
            raise Exception("Unsupported matrix length {} (must be 4, 9, or 16).".format(len(l)))


    def __getstate__(self):
        rv = { }

        for i in range(16):
            rv[fields[i]] = self.m[i]

        return rv

    def __setstate__(self, state):

        memset(self.m, 0, sizeof(float) * 16)

        self.zdz = 1.0
        self.wdw = 1.0

        for i in range(16):
            if fields[i] in state:
                self.m[i] = state[fields[i]]

    def __mul__(Matrix self, Matrix other):

        cdef Matrix rv = Matrix(None)

        rv.xdx = other.wdx*self.xdw + other.xdx*self.xdx + other.ydx*self.xdy + other.zdx*self.xdz
        rv.xdy = other.wdy*self.xdw + other.xdy*self.xdx + other.ydy*self.xdy + other.zdy*self.xdz
        rv.xdz = other.wdz*self.xdw + other.xdz*self.xdx + other.ydz*self.xdy + other.zdz*self.xdz
        rv.xdw = other.wdw*self.xdw + other.xdw*self.xdx + other.ydw*self.xdy + other.zdw*self.xdz

        rv.ydx = other.wdx*self.ydw + other.xdx*self.ydx + other.ydx*self.ydy + other.zdx*self.ydz
        rv.ydy = other.wdy*self.ydw + other.xdy*self.ydx + other.ydy*self.ydy + other.zdy*self.ydz
        rv.ydz = other.wdz*self.ydw + other.xdz*self.ydx + other.ydz*self.ydy + other.zdz*self.ydz
        rv.ydw = other.wdw*self.ydw + other.xdw*self.ydx + other.ydw*self.ydy + other.zdw*self.ydz

        rv.zdx = other.wdx*self.zdw + other.xdx*self.zdx + other.ydx*self.zdy + other.zdx*self.zdz
        rv.zdy = other.wdy*self.zdw + other.xdy*self.zdx + other.ydy*self.zdy + other.zdy*self.zdz
        rv.zdz = other.wdz*self.zdw + other.xdz*self.zdx + other.ydz*self.zdy + other.zdz*self.zdz
        rv.zdw = other.wdw*self.zdw + other.xdw*self.zdx + other.ydw*self.zdy + other.zdw*self.zdz

        rv.wdx = other.wdx*self.wdw + other.xdx*self.wdx + other.ydx*self.wdy + other.zdx*self.wdz
        rv.wdy = other.wdy*self.wdw + other.xdy*self.wdx + other.ydy*self.wdy + other.zdy*self.wdz
        rv.wdz = other.wdz*self.wdw + other.xdz*self.wdx + other.ydz*self.wdy + other.zdz*self.wdz
        rv.wdw = other.wdw*self.wdw + other.xdw*self.wdx + other.ydw*self.wdy + other.zdw*self.wdz

        return rv

    def __getitem__(Matrix self, int index):
        if 0 <= index < 16:
            return self.m[index]

        raise IndexError("Matrix index out of range.")

    def __setitem__(Matrix self, int index, float value):
        if 0 <= index < 16:
            self.m[index] = value
            return

        raise IndexError("Matrix index out of range.")

    def __repr__(Matrix self):
        cdef int x, y

        rv = "Matrix(["

        for 0 <= y < 4:
            if y:
                rv += "\n        "
            for 0 <= x < 4:
                rv += "{:8.5f}, ".format(self.m[x + y * 4])

        return rv + "])"

    def transform(Matrix self, float x, float y, float z=0.0, float w=1.0, int components=2):
        cdef float ox, oy, oz, ow

        self.transform4(&ox, &oy, &oz, &ow, x, y, z, w)

        if components == 2:
            return (ox, oy)
        elif components == 3:
            return (ox, oy, oz)
        elif components == 4:
            return (ox, oy, oz, ow)


    def __richcmp__(Matrix self, Matrix other, op):

        if op != 2:
            return NotImplemented

        if self is other:
            return True

        cdef int i
        cdef double total

        total = 0

        for 0 < i < 16:
            total += abs(self.m[i] - other.m[i])

        return total < .0001

    cpdef bint is_unit_aligned(Matrix self):
        """
        Returns true if exactly one of abs(xdx) or abs(xdy) is 1.0, and
        the same for xdy and ydy. This is intended to report if a matrix
        is aligned to the axes.
        """

        cdef int i
        cdef float v
        cdef float total_1
        cdef float total_2

        total_1 = 0
        total_2 = 0

        for 0 < i < 16:
            v = abs(self.m[i])
            total_1 += abs(v - aligned_1[i])
            total_2 += abs(v - aligned_2[i])

        return (total_1 < .0001) or (total_2 < .0001)

cdef class Matrix2D(Matrix):

    def __init__(Matrix2D self, double xdx, double xdy, double ydx, double ydy):
        memset(self.m, 0, sizeof(float) * 16)

        self.xdx = xdx
        self.xdy = xdy
        self.ydx = ydx
        self.ydy = ydy

        self.zdz = 1.0
        self.wdw = 1.0


include "matrix_functions.pxi"



#@PydevCodeAnalysisIgnore
"""Easy to use dialogs.

Message(msg) -- display a message and an OK button.
AskString(prompt, default) -- ask for a string, display OK and Cancel buttons.
AskPassword(prompt, default) -- like AskString(), but shows text as bullets.
AskYesNoCancel(question, default) -- display a question and Yes, No and Cancel buttons.
GetArgv(optionlist, commandlist) -- fill a sys.argv-like list using a dialog
AskFileForOpen(...) -- Ask the user for an existing file
AskFileForSave(...) -- Ask the user for an output file
AskFolder(...) -- Ask the user to select a folder
bar = Progress(label, maxvalue) -- Display a progress bar
bar.set(value) -- Set value
bar.inc( *amount ) -- increment value by amount (default=1)
bar.label( *newlabel ) -- get or set text label.

More documentation in each function.
This module uses DLOG resources 260 and on.
Based upon STDWIN dialogs with the same names and functions.
"""

from __future__ import division
from __future__ import print_function

import os

import ctypes
import ctypes.wintypes as wintypes

from EasyDialogsResources import resources


__all__ = ['Message', 'AskString', 'AskPassword', 'AskYesNoCancel',
    'GetArgv', 'AskFileForOpen', 'AskFileForSave', 'AskFolder',
    'ProgressBar']


# Load required Windows DLLs
comdlg32 = ctypes.windll.comdlg32
kernel32 = ctypes.windll.kernel32
ole32 = ctypes.windll.ole32
shell32 = ctypes.windll.shell32
user32 = ctypes.windll.user32


# Windows Constants
BFFM_INITIALIZED = 1
BFFM_SETOKTEXT = 1129
BFFM_SETSELECTIONA = 1126
BIF_EDITBOX = 16
BS_DEFPUSHBUTTON = 1
CB_ADDSTRING = 323
CB_GETCURSEL = 327
CB_SETCURSEL = 334
CDM_SETCONTROLTEXT = 1128
EM_GETLINECOUNT = 186
EM_GETMARGINS = 212
EM_POSFROMCHAR = 214
EM_SETSEL = 177
GWL_STYLE = -16
IDC_STATIC = -1
IDCANCEL = 2
IDNO = 7
IDOK = 1
IDYES = 6
MAX_PATH = 260
OFN_ALLOWMULTISELECT = 512
OFN_ENABLEHOOK = 32
OFN_ENABLESIZING = 8388608
OFN_ENABLETEMPLATEHANDLE = 128
OFN_EXPLORER = 524288
OFN_OVERWRITEPROMPT = 2
OPENFILENAME_SIZE_VERSION_400 = 76
PBM_GETPOS = 1032
PBM_SETMARQUEE = 1034
PBM_SETPOS = 1026
PBM_SETRANGE = 1025
PBM_SETRANGE32 = 1030
PBS_MARQUEE = 8
PM_REMOVE = 1
SW_HIDE = 0
SW_SHOW = 5
SW_SHOWNORMAL = 1
SWP_NOACTIVATE = 16
SWP_NOMOVE = 2
SWP_NOSIZE = 1
SWP_NOZORDER = 4
VER_PLATFORM_WIN32_NT = 2
WM_COMMAND = 273
WM_GETTEXT = 13
WM_GETTEXTLENGTH = 14
WM_INITDIALOG = 272
WM_NOTIFY = 78


# Windows function prototypes
BrowseCallbackProc = ctypes.WINFUNCTYPE(ctypes.c_int, wintypes.HWND, ctypes.c_uint, wintypes.LPARAM, wintypes.LPARAM)
DialogProc = ctypes.WINFUNCTYPE(ctypes.c_int, wintypes.HWND, ctypes.c_uint, wintypes.WPARAM, wintypes.LPARAM)
EnumChildProc = ctypes.WINFUNCTYPE(ctypes.c_int, wintypes.HWND, wintypes.LPARAM)
LPOFNHOOKPROC = ctypes.WINFUNCTYPE(ctypes.c_int, wintypes.HWND, ctypes.c_uint, wintypes.WPARAM, wintypes.LPARAM)


# Windows types
LPCTSTR = ctypes.c_char_p
LPTSTR = ctypes.c_char_p
LPVOID = ctypes.c_voidp
TCHAR = ctypes.c_char

class OSVERSIONINFO(ctypes.Structure):
    _fields_ = [
                ("dwOSVersionInfoSize", wintypes.DWORD),
                ("dwMajorVersion", wintypes.DWORD),
                ("dwMinorVersion", wintypes.DWORD),
                ("dwBuildNumber", wintypes.DWORD),
                ("dwPlatformId", wintypes.DWORD),
                ("szCSDVersion", TCHAR*128)
               ]
    def __init__(self):
        ctypes.Structure.__init__(self)
        self.dwOSVersionInfoSize = ctypes.sizeof(self)

class OPENFILENAME(ctypes.Structure):
    _fields_ = [
                ("lStructSize", wintypes.DWORD),
                ("hwndOwner", wintypes.HWND),
                ("hInstance", LPCTSTR),
                ("lpstrFilter", LPCTSTR),
                ("lpstrCustomFilter", LPTSTR),
                ("nMaxCustFilter", wintypes.DWORD),
                ("nFilterIndex", wintypes.DWORD),
                ("lpstrFile", LPTSTR),
                ("nMaxFile", wintypes.DWORD),
                ("lpstrFileTitle", LPTSTR),
                ("nMaxFileTitle", wintypes.DWORD),
                ("lpstrInitialDir", LPCTSTR),
                ("lpstrTitle", LPCTSTR),
                ("flags", wintypes.DWORD),
                ("nFileOffset", wintypes.WORD),
                ("nFileExtension", wintypes.WORD),
                ("lpstrDefExt", LPCTSTR),
                ("lCustData", wintypes.LPARAM),
                ("lpfnHook", LPOFNHOOKPROC),
                ("lpTemplateName", LPCTSTR),
                ("pvReserved", LPVOID),
                ("dwReserved", wintypes.DWORD),
                ("FlagsEx", wintypes.DWORD)
               ]

    def __init__(self):
        ctypes.Structure.__init__(self)
        versionInfo = OSVERSIONINFO()
        kernel32.GetVersionExA(ctypes.byref(versionInfo))
        if versionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT:
            if (versionInfo.dwMajorVersion, versionInfo.dwMinorVersion) == (4, 0):
                self.lStructSize = OPENFILENAME_SIZE_VERSION_400
            elif versionInfo.dwMajorVersion >= 5:
                self.lStructSize = ctypes.sizeof(OPENFILENAME)
            else:
                raise RuntimeError('Windows NT 4.0 or later is required.')
        else:
            self.lStructSize = ctypes.sizeof(OPENFILENAME) - 12

class BROWSEINFO(ctypes.Structure):
    _fields_ = [
                ("hwndOwner", wintypes.HWND),
                ("pidlRoot", LPVOID),
                ("pszDisplayName", LPTSTR),
                ("lpszTitle", LPCTSTR),
                ("ulFlags", ctypes.c_uint),
                ("lpfn", BrowseCallbackProc),
                ("lParam", wintypes.LPARAM),
                ("iImage", ctypes.c_int)
               ]


# Windows macro emulation
def HIWORD(x):
    return (x >> 16) & 0xffff

def LOWORD(x):
    return x & 0xffff

def MAKELONG(a, b):
    return (a & 0xffff) | ((b & 0xffff) << 16)

MAKELPARAM = MAKELONG


# Utilities
def CenterWindow(hwnd):
    desktopRect = GetWindowRect(user32.GetDesktopWindow())
    myRect = GetWindowRect(hwnd)
    x = width(desktopRect) // 2 - width(myRect) // 2
    y = height(desktopRect) // 2 - height(myRect) // 2
    user32.SetWindowPos(hwnd, 0,
                        desktopRect.left + x,
                        desktopRect.top + y,
                        0, 0,
                        SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER
                       )

def EnumChildWindows(hwnd):
    childWindows = []
    def enumChildProc(hwnd, lParam):
        childWindows.append(hwnd)
        return True
    enumChildProc = EnumChildProc(enumChildProc)
    user32.EnumChildWindows(hwnd, enumChildProc, 0)
    return childWindows

def GetText(hwnd):
    length = user32.SendMessageA(hwnd, WM_GETTEXTLENGTH, 0, 0) + 1
    buffer = ctypes.c_char_p('\0' * length)
    user32.SendMessageA(hwnd, WM_GETTEXT, length, buffer)
    return buffer.value

def GetWindowRect(hwnd):
    rect = wintypes.RECT()
    user32.GetWindowRect(hwnd, ctypes.byref(rect))
    return rect

def width(rect):
    return rect.right-rect.left

def height(rect):
    return rect.bottom-rect.top


# Mac API emulation
def AutoSizeDialog(dialog, center=True):
    classNameBuffer = ctypes.c_buffer(1024)
    children = []
    editBoxes = []
    for child in EnumChildWindows(dialog):
        children.append(child)
        user32.GetClassNameA(child, classNameBuffer, 1024)
        if classNameBuffer.value.lower() == 'edit':
            editBoxes.append(child)

    editBoxes.sort(lambda x, y: cmp(GetWindowRect(x).top, GetWindowRect(y).top))

    for editBox in editBoxes:
        text = GetText(editBox)
        user32.SetWindowTextA(editBox, 'X\r\nX')
        if user32.SendMessageA(editBox, EM_GETLINECOUNT, 0, 0) < 2:
            user32.SetWindowTextA(editBox, text)
        else:
            lineHeight = HIWORD(user32.SendMessageA(editBox, EM_POSFROMCHAR, 3, 0)) - HIWORD(user32.SendMessageA(editBox, EM_POSFROMCHAR, 0, 0))
            user32.SetWindowTextA(editBox, text)
            newHeight = user32.SendMessageA(editBox, EM_GETLINECOUNT, 0, 0)*lineHeight + 2*LOWORD(user32.SendMessageA(editBox, EM_GETMARGINS, 0, 0))
            editBoxRect = GetWindowRect(editBox)
            oldHeight = height(editBoxRect)
            if newHeight != oldHeight:
                topLeft = wintypes.POINT(editBoxRect.left, editBoxRect.top)
                user32.ScreenToClient(dialog, ctypes.byref(topLeft))
                user32.MoveWindow(editBox, topLeft.x, topLeft.y, width(editBoxRect), newHeight, False)

                for child in children:
                    childRect = GetWindowRect(child)
                    if childRect.top >= editBoxRect.bottom:
                        topLeft = wintypes.POINT(childRect.left, childRect.top)
                        user32.ScreenToClient(dialog, ctypes.byref(topLeft))
                        user32.MoveWindow(child,
                                          topLeft.x, topLeft.y + newHeight - oldHeight,
                                          width(childRect), height(childRect),
                                          False
                                         )

                dRect = GetWindowRect(dialog)
                dRect.bottom += newHeight - oldHeight
                topLeft = wintypes.POINT(dRect.left, dRect.top)
                user32.MoveWindow(dialog, topLeft.x, topLeft.y, width(dRect), height(dRect), False)
                user32.InvalidateRect(dialog, None, True)

    if center:
        CenterWindow(dialog)


# Mac EasyDialogs emulation
def crlf2lf(text):
    if '\r\n' in text:
        text = '\n'.join(text.split('\r\n'))
    return text


def lf2crlf(text):
    if '\n' in text:
        text = '\r\n'.join(text.split('\n'))
    if len(text) > 253:
        text = text[:253] + '\205'
    return text


def Message(msg, id=260, ok=None):
    """Display a MESSAGE string.

    Return when the user clicks the OK button or presses Return.

    The MESSAGE string can be at most 255 characters long.
    """
    def DlgProc(hwnd, uMsg, wParam, lParam):
        if uMsg == WM_INITDIALOG:
            user32.SetWindowTextA(user32.GetDlgItem(hwnd, 1002), lf2crlf(msg))
            if ok:
                user32.SetWindowTextA(user32.GetDlgItem(hwnd, IDOK), ok)
            AutoSizeDialog(hwnd)
            return False
        if uMsg == WM_COMMAND and LOWORD(wParam) == IDOK:
            user32.EndDialog(hwnd, IDOK)
            return True
        if uMsg == WM_COMMAND and LOWORD(wParam) == IDCANCEL:
            user32.EndDialog(hwnd, IDCANCEL)
            return True
        else:
            return False

    # This next line is needed to prevent gc of the callback
    myDialogProc = DialogProc(DlgProc)
    return user32.DialogBoxIndirectParamA(0, resources[id], 0, myDialogProc, 0)


def AskString(prompt, default = '', id=261, ok=None, cancel=None):
    """Display a PROMPT string and a text entry field with a DEFAULT string.

    Return the contents of the text entry field when the user clicks the
    OK button or presses Return.
    Return None when the user clicks the Cancel button.

    If omitted, DEFAULT is empty.

    The PROMPT and DEFAULT strings, as well as the return value,
    can be at most 255 characters long.
    """
    result = [None]
    def DlgProc(hwnd, uMsg, wParam, lParam):
        if uMsg == WM_INITDIALOG:
            user32.SetWindowTextA(user32.GetDlgItem(hwnd, 1003), lf2crlf(prompt))
            user32.SetWindowTextA(user32.GetDlgItem(hwnd, 1004), lf2crlf(default))
            if ok:
                user32.SetWindowTextA(user32.GetDlgItem(hwnd, IDOK), ok)
            if cancel:
                user32.SetWindowTextA(user32.GetDlgItem(hwnd, IDCANCEL), cancel)
            AutoSizeDialog(hwnd)
            return False
        if uMsg == WM_COMMAND and LOWORD(wParam) == IDOK:
            result[0] = crlf2lf(GetText(user32.GetDlgItem(hwnd, 1004)))
            user32.EndDialog(hwnd, IDOK)
            return True
        if uMsg == WM_COMMAND and LOWORD(wParam) == IDCANCEL:
            user32.EndDialog(hwnd, IDCANCEL)
            return True
        else:
            return False

    # This next line is needed to prevent gc of the callback
    myDialogProc = DialogProc(DlgProc)
    user32.DialogBoxIndirectParamA(0, resources[id], 0, myDialogProc, 0)
    return result[0]


def AskPassword(prompt, default='', id=264, ok=None, cancel=None):
    """Display a PROMPT string and a text entry field with a DEFAULT string.
    The string is displayed as bullets only.

    Return the contents of the text entry field when the user clicks the
    OK button or presses Return.
    Return None when the user clicks the Cancel button.

    If omitted, DEFAULT is empty.

    The PROMPT and DEFAULT strings, as well as the return value,
    can be at most 255 characters long.
    """
    return AskString(prompt, default=default, id=id, ok=ok, cancel=cancel)


def AskYesNoCancel(question, default=0, yes=None, no=None, cancel=None, id=262):
    """Display a QUESTION string which can be answered with Yes or No.

    Return 1 when the user clicks the Yes button.
    Return 0 when the user clicks the No button.
    Return -1 when the user clicks the Cancel button.

    When the user presses Return, the DEFAULT value is returned.
    If omitted, this is 0 (No).

    The QUESTION string can be at most 255 characters.
    """
    def DlgProc(hwnd, uMsg, wParam, lParam):
        if uMsg == WM_INITDIALOG:
            user32.SetWindowTextA(user32.GetDlgItem(hwnd, 1005), lf2crlf(question))

            if yes is not None:
                button = user32.GetDlgItem(hwnd, IDYES)
                if yes == '':
                    user32.ShowWindow(button, False)
                else:
                    user32.SetWindowTextA(button, yes)
            if no is not None:
                button = user32.GetDlgItem(hwnd, IDNO)
                if no == '':
                    user32.ShowWindow(button, False)
                else:
                    user32.SetWindowTextA(button, no)
            if cancel is not None:
                button = user32.GetDlgItem(hwnd, IDCANCEL)
                if cancel == '':
                    user32.ShowWindow(button, False)
                else:
                    user32.SetWindowTextA(button, cancel)

            button = None
            if default == 1:
                button = user32.GetDlgItem(hwnd, IDYES)
            elif default == 0:
                button = user32.GetDlgItem(hwnd, IDNO)
            elif default == -1:
                button = user32.GetDlgItem(hwnd, IDCANCEL)
            if button:
                style = user32.GetWindowLongA(button, GWL_STYLE)
                user32.SetWindowLongA(button, GWL_STYLE, style | BS_DEFPUSHBUTTON)
                user32.SetFocus(button)

            AutoSizeDialog(hwnd)
            return False
        if uMsg == WM_COMMAND and LOWORD(wParam) == IDYES:
            user32.EndDialog(hwnd, 1)
            return True
        elif uMsg == WM_COMMAND and LOWORD(wParam) == IDNO:
            user32.EndDialog(hwnd, 0)
            return True
        elif uMsg == WM_COMMAND and LOWORD(wParam) == IDCANCEL:
            user32.EndDialog(hwnd, -1)
            return True
        else:
            return False

    # This next line is needed to prevent gc of the callback
    myDialogProc = DialogProc(DlgProc)
    return user32.DialogBoxIndirectParamA(0, resources[id], 0, myDialogProc, 0)


class ProgressBar:
    def __init__(self, title="Working...", maxval=0, label="", id=263):
        self._label = lf2crlf(label)
        def DlgProc(hwnd, uMsg, wParam, lParam):
            if uMsg == WM_INITDIALOG:
                user32.SetWindowTextA(hwnd, title)
                user32.SetWindowTextA(user32.GetDlgItem(hwnd, 1002), DlgProc.label)
            if uMsg == WM_COMMAND and LOWORD(wParam) == IDOK:
                user32.EndDialog(hwnd, IDOK)
                return True
            if uMsg == WM_COMMAND and LOWORD(wParam) == IDCANCEL:
                user32.EndDialog(hwnd, IDCANCEL)
                return True
            else:
                return False

        # Why a function attribute? Originally nested scopes
        # were used to get at self._label. This caused a cyclic reference
        # and since ProgressBar has a __del__ method (which closes the
        # window) it would never be collected (and hence the window
        # wouldn't close). Think of label as being like self. When this
        # function it used it will be used as a bound method where the
        # label is the object.
        DlgProc.label = self._label

        # This next line is needed to prevent gc of the callback
        self.wrappedDlgProc = DialogProc(DlgProc)
        self.hwnd = user32.CreateDialogIndirectParamA(0, resources[id], 0, self.wrappedDlgProc, 0)

        AutoSizeDialog(self.hwnd)
        user32.ShowWindow(self.hwnd, SW_SHOWNORMAL)
        self.set(0, maxval)

    def __del__(self):
        user32.DestroyWindow(self.hwnd)

    def title(self, newstr=""):
        """title(text) - Set title of progress window"""
        user32.SetWindowTextA(self.hwnd, newstr)
        self._pump()

    def label( self, *newstr ):
        """label(text) - Set text in progress box"""
        if newstr:
            self._label = lf2crlf(newstr[0])
        user32.SetWindowTextA(user32.GetDlgItem(self.hwnd, 1002), self._label)
        AutoSizeDialog(self.hwnd, center=False)
        self._pump()


    def _pump(self):
        msg = wintypes.MSG()
        while user32.PeekMessageA(ctypes.byref(msg), 0, 0, 0, PM_REMOVE):
            if not user32.IsDialogMessage(self.hwnd, ctypes.byref(msg)):
                user32.TranslateMessage(ctypes.byref(msg))
                user32.DispatchMessageA(ctypes.byref(msg))

        if not user32.IsWindowVisible(self.hwnd):
            raise KeyboardInterrupt


    def _update(self, value):
        maxval = self.maxval
        progbar = user32.GetDlgItem(self.hwnd, 1003)
        if maxval == 0:     # an indeterminate bar
            # make the bar grow and wrap around in case Marquee bars aren't supported
            pass
            pos = user32.SendMessageA(progbar, PBM_GETPOS, 0, 0)
            user32.SendMessageA(progbar, PBM_SETRANGE, 0, MAKELPARAM(0, 10))
            user32.SendMessageA(progbar, PBM_SETPOS, (pos+1) % 10, 0)
        else:               # a determinate bar
            if maxval > 32767:
                value = int(value/(maxval/32767.0))
                maxval = 32767
            maxval = int(maxval)
            value = int(value)
            if maxval > 65535:
                user32.SendMessageA(progbar, PBM_SETRANGE32, 0, maxval)
            else:
                user32.SendMessageA(progbar, PBM_SETRANGE, 0, MAKELPARAM(0, maxval))
            user32.SendMessageA(progbar, PBM_SETPOS, value, 0)  # set the bar length
        self._pump()


    def set(self, value, max=None):
        """set(value) - Set progress bar position"""
        if max != None:
            self.maxval = max
            bar = user32.GetDlgItem(self.hwnd, 1003)
            if max <= 0:    # indeterminate bar
                self.maxval = 0
                style = user32.GetWindowLongA(bar, GWL_STYLE)
                if not (style & PBS_MARQUEE):
                    user32.SetWindowLongA(bar, GWL_STYLE, style | PBS_MARQUEE)
                    user32.SendMessageA(bar, PBM_SETMARQUEE, 1, 10)
            else:           # determinate bar
                user32.SendMessageA(bar, PBM_SETMARQUEE, 0, 0)
                style = user32.GetWindowLongA(bar, GWL_STYLE)
                if style & PBS_MARQUEE:
                    user32.SetWindowLongA(bar, GWL_STYLE, style - PBS_MARQUEE)
        if value < 0:
            value = 0
        elif value > self.maxval:
            value = self.maxval
        self.curval = value
        self._update(value)

    def inc(self, n=1):
        """inc(amt) - Increment progress bar position"""
        self.set(self.curval + n)


def AskFileForOpen(
        message=None,
        typeList=None,
        # From here on the order is not documented
        version=None,
        defaultLocation=None,
        dialogOptionFlags=None,
        location=None,
        clientName=None,
        windowTitle=None,
        actionButtonLabel=None,
        cancelButtonLabel=None,
        preferenceKey=None,
        popupExtension=None,
        eventProc=None,
        previewProc=None,
        filterProc=None,
        wanted=None,
        multiple=None,
        # the following are for implementation use only
        defaultfn='',
        defaultext=None,
        fn=comdlg32.GetOpenFileNameA):
    """Display a dialog asking the user for a file to open.

    wanted is the return type wanted: FSSpec, FSRef, unicode or string (default)
    the other arguments can be looked up in Apple's Navigation Services documentation

    Windows differences:

    typeList is used for the same purpose, but file type handling is different
    between Windows and Mac, so the form of this argument is different. In an
    attempt to remain as similar as possible, a list of extensions can be
    supplied (e.g., ['*', 'txt', 'bat']). A more complete form is also
    allowed: [('All Files (*.*)', '*.*'), ('C Files (*.c, *.h)', '*.c;*.h')].
    The first item in each tuple is the text description presented to the
    user. The second item in each tuple is a semi-colon seperated list of
    standard Windows wildcard patterns that will match files described in the
    text description.

    The folowing parameters are ignored on Windows:
    clientName, dialogOptionFlags, eventProc, filterProc, popupExtension,
    preferenceKey, previewProc, version, wanted
    """

    ofn = OPENFILENAME()
    filename = defaultfn + '\0'*1024
    ofn.lpstrFile = filename
    ofn.nMaxFile = len(filename)
    ofn.flags = OFN_ENABLEHOOK | OFN_EXPLORER | OFN_ENABLESIZING | OFN_OVERWRITEPROMPT
    if multiple:
        ofn.flags |= OFN_ALLOWMULTISELECT
    ofn.lpstrTitle = windowTitle
    ofn.lpstrInitialDir = defaultLocation

    if typeList and filter(None, typeList):
        lpstrFilter = ''
        for typeSpec in typeList:
            try:
                lpstrFilter += '*.'+typeSpec+'\0*.'+typeSpec+'\0'
                if defaultext and not ofn.lpstrDefExt:
                    ofn.lpstrDefExt = typeSpec
            except TypeError:
                lpstrFilter += '%s\0%s\0' % typeSpec
                if defaultext and not ofn.lpstrDefExt:
                    ext = os.path.splitext(typeSpec[1])[1].lstrip('.')
                    if ext != '*':
                        ofn.lpstrDefExt = ext
        ofn.lpstrFilter = lpstrFilter + '\0'
    else:
        ofn.lpstrFilter = 'All Files (*.*)\0*.*\0\0'

    if message:
        ofn.flags |= OFN_ENABLETEMPLATEHANDLE
        ofn.hInstance = resources[270]

    hookProcInitDone = []
    def hookProc(hdlg, uiMsg, wParam, lParam):
        if uiMsg == WM_INITDIALOG:
            dialog = user32.GetParent(hdlg)
            if message:
                handle = user32.FindWindowExA(hdlg, 0, 0, 0)
                user32.SetWindowTextA(handle, message)
            if actionButtonLabel:
                user32.SendMessageA(dialog, CDM_SETCONTROLTEXT, IDOK, actionButtonLabel)
            if cancelButtonLabel:
                user32.SendMessageA(dialog, CDM_SETCONTROLTEXT, IDCANCEL, cancelButtonLabel)
        elif uiMsg == WM_NOTIFY:
            if not hookProcInitDone:
                hookProcInitDone.append(True)
                dialog = user32.GetParent(hdlg)
                desktop = user32.GetDesktopWindow()
                if message:
                    clientRect = wintypes.RECT()
                    user32.GetClientRect(dialog, ctypes.byref(clientRect))
                    dlgRect = wintypes.RECT()
                    user32.GetWindowRect(hdlg, ctypes.byref(dlgRect))
                    user32.SetWindowPos(hdlg, 0,
                                      0, 0,
                                      width(clientRect), height(dlgRect),
                                      SWP_NOMOVE | SWP_NOZORDER)
                    handle = user32.GetDlgItem(hdlg, IDC_STATIC)
                    user32.SetWindowPos(handle, 0,
                                      0, 0,
                                      width(clientRect), height(dlgRect),
                                      SWP_NOMOVE | SWP_NOZORDER)
                if location:
                    x, y = location
                    desktopRect = wintypes.RECT()
                    user32.GetWindowRect(desktop, ctypes.byref(desktopRect))
                    user32.SetWindowPos(dialog, 0,
                                      desktopRect.left + x,
                                      desktopRect.top + y, 0, 0,
                                      SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER)
                else: # center
                    CenterWindow(dialog)
        return 0

    ofn.lpfnHook = LPOFNHOOKPROC(hookProc)

    if fn(ctypes.byref(ofn)):
        filenames = filter(None, filename.split('\0'))
        if len(filenames) > 1:
            dir, filenames = filenames[0], filenames[1:]
            return map(lambda fn: os.path.join(dir, fn), filenames)
        elif multiple:
            return filenames
        else:
            return filenames[0]
    else:
        return None


def AskFileForSave(
        message=None,
        savedFileName=None,
        # From here on the order is not documented
        version=None,
        defaultLocation=None,
        dialogOptionFlags=None,
        location=None,
        clientName=None,
        windowTitle=None,
        actionButtonLabel=None,
        cancelButtonLabel=None,
        preferenceKey=None,
        popupExtension=None,
        eventProc=None,
        fileType=None,
        fileCreator=None,
        wanted=None,
        multiple=None):
    """Display a dialog asking the user for a filename to save to.

    wanted is the return type wanted: FSSpec, FSRef, unicode or string (default)
    the other arguments can be looked up in Apple's Navigation Services documentation

    Windows differences:

    fileType is used for the same purpose, but file type handling is different
    between Windows and Mac, so the form of this argument is different. In an
    attempt to remain as similar as possible, an extensions can be supplied
    (e.g., 'txt'). A more complete form is also	allowed:
    ('Text Files (*.txt)', '*.txt'). The first item in the tuple is the text
    description presented to the user. The second item in the tuple is a
    standard Windows wildcard pattern that will match files described in the
    text description.

    The folowing parameters are ignored on Windows:
    clientName, dialogOptionFlags, eventProc, fileCreator, filterProc,
    multiple, popupExtension, preferenceKey, previewProc, version, wanted
    """
    return AskFileForOpen(
                        message=message, typeList=[fileType], version=version,
                        defaultLocation=defaultLocation,
                        dialogOptionFlags=dialogOptionFlags, location=location,
                        clientName=clientName, windowTitle=windowTitle,
                        actionButtonLabel=actionButtonLabel,
                        cancelButtonLabel=cancelButtonLabel,
                        preferenceKey=preferenceKey,
                        popupExtension=popupExtension, eventProc=eventProc,
                        wanted=wanted, multiple=multiple,
                        defaultfn=savedFileName or '',
                        defaultext=True,
                        fn=comdlg32.GetSaveFileNameA)


def AskFolder(
        message=None,
        # From here on the order is not documented
        version=None,
        defaultLocation=None,
        dialogOptionFlags=None,
        location=None,
        clientName=None,
        windowTitle=None,
        actionButtonLabel=None,
        cancelButtonLabel=None,
        preferenceKey=None,
        popupExtension=None,
        eventProc=None,
        filterProc=None,
        wanted=None,
        multiple=None):
    """Display a dialog asking the user for select a folder.

    wanted is the return type wanted: FSSpec, FSRef, unicode or string (default)
    the other arguments can be looked up in Apple's Navigation Services documentation

    Windows differences:

    The folowing parameters are ignored on Windows:
    clientName, dialogOptionFlags, eventProc, filterProc,
    multiple, popupExtension, preferenceKey, version, wanted
    """


    def BrowseCallback(hwnd, uMsg, lParam, lpData):
        if uMsg == BFFM_INITIALIZED:
            if actionButtonLabel:
                label = unicode(actionButtonLabel, errors='replace')
                user32.SendMessageW(hwnd, BFFM_SETOKTEXT, 0, label)
            if cancelButtonLabel:
                cancelButton = user32.GetDlgItem(hwnd, IDCANCEL)
                if cancelButton:
                    user32.SetWindowTextA(cancelButton, cancelButtonLabel)
            if windowTitle:
                user32.SetWindowTextA(hwnd, windowTitle)
            if defaultLocation:
                user32.SendMessageW(hwnd, BFFM_SETSELECTIONA, 1, defaultLocation.replace('/', '\\'))
            if location:
                x, y = location
                desktopRect = wintypes.RECT()
                user32.GetWindowRect(0, ctypes.byref(desktopRect))
                user32.SetWindowPos(hwnd, 0,
                                  desktopRect.left + x,
                                  desktopRect.top + y, 0, 0,
                                  SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER)
            else:
                CenterWindow(hwnd)
        return 0

    # This next line is needed to prevent gc of the callback
    callback = BrowseCallbackProc(BrowseCallback)

    browseInfo = BROWSEINFO()
    browseInfo.pszDisplayName = ctypes.c_char_p('\0' * (MAX_PATH+1))
    browseInfo.lpszTitle = message
    browseInfo.lpfn = callback
    #~ browseInfo.ulFlags = BIF_EDITBOX

    # BIF_NEWDIALOGSTYLE is not used because it only works with apartment
    # threading. Using this style with COINIT_MULTITHREADED can cause problems.
    pidl = shell32.SHBrowseForFolder(ctypes.byref(browseInfo))
    if not pidl:
        result = None
    else:
        path = ctypes.c_char_p('\0' * (MAX_PATH+1))
        shell32.SHGetPathFromIDList(pidl, path)
        ole32.CoTaskMemFree(pidl)
        result = path.value
    return result


ARGV_ID=265
ARGV_ITEM_OK=1
ARGV_ITEM_CANCEL=2
ARGV_OPTION_GROUP=3
ARGV_OPTION_EXPLAIN=4
ARGV_OPTION_VALUE=5
ARGV_OPTION_ADD=6
ARGV_COMMAND_GROUP=7
ARGV_COMMAND_EXPLAIN=8
ARGV_COMMAND_ADD=9
ARGV_ADD_OLDFILE=10
ARGV_ADD_NEWFILE=11
ARGV_ADD_FOLDER=12
ARGV_CMDLINE_GROUP=13
ARGV_CMDLINE_DATA=14

def _setmenu(control, items):
    for item in items:
        if type(item) == type(()):
            label = item[0]
        else:
            label = item
        if label[-1] == '=' or label[-1] == ':':
            label = label[:-1]
        user32.SendMessageA(control, CB_ADDSTRING, 0, label)
    user32.SendMessageA(control, CB_SETCURSEL, 0, 0)

def _selectoption(d, optionlist, idx):
    if idx < 0 or idx >= len(optionlist):
        user32.MessageBeep(-1)
        return
    option = optionlist[idx]
    if type(option) == type(()):
        if len(option) == 4:
            help = option[2]
        elif len(option) > 1:
            help = option[-1]
        else:
            help = ''
    else:
        help = ''
    h = user32.GetDlgItem(d, ARGV_OPTION_EXPLAIN)
    if help and len(help) > 250:
        help = help[:250] + '...'
    user32.SetWindowTextA(h, help)
    hasvalue = 0
    if type(option) == type(()):
        label = option[0]
    else:
        label = option
    if label[-1] == '=' or label[-1] == ':':
        hasvalue = 1
    h = user32.GetDlgItem(d, ARGV_OPTION_VALUE)
    user32.SetWindowTextA(h, '')
    if hasvalue:
        user32.ShowWindow(h, SW_SHOW)
        #~ d.SelectDialogItemText(ARGV_OPTION_VALUE, 0, 0)
    else:
        user32.ShowWindow(h, SW_HIDE)

def GetArgv(optionlist=None, commandlist=None, addoldfile=1, addnewfile=1, addfolder=1, id=ARGV_ID):
    commandLineContents = []
    def DlgProc(hwnd, uMsg, wParam, lParam):
        if uMsg == WM_INITDIALOG:
            if optionlist:
                _setmenu(user32.GetDlgItem(hwnd, ARGV_OPTION_GROUP), optionlist)
                _selectoption(hwnd, optionlist, 0)
            else:
                for id in (30, ARGV_OPTION_GROUP, ARGV_OPTION_VALUE, ARGV_OPTION_ADD):
                    user32.EnableWindow(user32.GetDlgItem(hwnd, id), False)
            if commandlist:
                _setmenu(user32.GetDlgItem(hwnd, ARGV_COMMAND_GROUP), commandlist)
                if type(commandlist[0]) == type(()) and len(commandlist[0]) > 1:
                    help = commandlist[0][-1]
                    h = user32.GetDlgItem(hwnd, ARGV_COMMAND_EXPLAIN)
                    user32.SetWindowTextA(h, help)
            else:
                for id in (70, ARGV_COMMAND_GROUP, ARGV_COMMAND_ADD):
                    user32.EnableWindow(user32.GetDlgItem(hwnd, id), False)
            if not addoldfile:
                user32.EnableWindow(user32.GetDlgItem(hwnd, ARGV_ADD_OLDFILE), False)
            if not addnewfile:
                user32.EnableWindow(user32.GetDlgItem(hwnd, ARGV_ADD_NEWFILE), False)
            if not addfolder:
                user32.EnableWindow(user32.GetDlgItem(hwnd, ARGV_ADD_FOLDER), False)
            CenterWindow(hwnd)
            return False
        elif uMsg == WM_COMMAND and LOWORD(wParam) == IDOK:
            h = user32.GetDlgItem(hwnd, ARGV_CMDLINE_DATA)
            commandLineContents.append(GetText(h))
            user32.EndDialog(hwnd, IDOK)
            return True
        elif uMsg == WM_COMMAND and LOWORD(wParam) == IDCANCEL:
            user32.EndDialog(hwnd, IDCANCEL)
            return True

        if uMsg == WM_COMMAND:
            stringstoadd = []
            n = LOWORD(wParam)
            if n == ARGV_OPTION_GROUP:
                idx = user32.SendMessageA(lParam, CB_GETCURSEL, 0, 0)
                _selectoption(hwnd, optionlist, idx)
                return True
            elif n == ARGV_OPTION_ADD:
                idx = user32.SendMessageA(user32.GetDlgItem(hwnd, ARGV_OPTION_GROUP), CB_GETCURSEL, 0, 0)
                if 0 <= idx < len(optionlist):
                    option = optionlist[idx]
                    if type(option) == type(()):
                        option = option[0]
                    if option[-1] == '=' or option[-1] == ':':
                        option = option[:-1]
                        h = user32.GetDlgItem(hwnd, ARGV_OPTION_VALUE)
                        value = GetText(h)
                    else:
                        value = ''
                    if len(option) == 1:
                        stringtoadd = '-' + option
                    else:
                        stringtoadd = '--' + option
                    stringstoadd[:] = [stringtoadd]
                    if value:
                        stringstoadd.append(value)
                else:
                    user32.MessageBeep(-1)
            elif n == ARGV_COMMAND_GROUP:
                idx = user32.SendMessageA(lParam, CB_GETCURSEL, 0, 0)
                if 0 <= idx < len(commandlist) and type(commandlist[idx]) == type(()) and \
                        len(commandlist[idx]) > 1:
                    help = commandlist[idx][-1]
                    h = user32.GetDlgItem(hwnd, ARGV_COMMAND_EXPLAIN)
                    user32.SetWindowTextA(h, help)
            elif n == ARGV_COMMAND_ADD:
                idx = user32.SendMessageA(user32.GetDlgItem(hwnd, ARGV_COMMAND_GROUP), CB_GETCURSEL, 0, 0)
                if 0 <= idx < len(commandlist):
                    command = commandlist[idx]
                    if type(command) == type(()):
                        command = command[0]
                    stringstoadd = [command]
                else:
                    user32.MessageBeep(-1)
            elif n == ARGV_ADD_OLDFILE:
                pathname = AskFileForOpen()
                if pathname:
                    stringstoadd = [pathname]
            elif n == ARGV_ADD_NEWFILE:
                pathname = AskFileForSave()
                if pathname:
                    stringstoadd = [pathname]
            elif n == ARGV_ADD_FOLDER:
                pathname = AskFolder()
                if pathname:
                    stringstoadd = [pathname]

            for stringtoadd in stringstoadd:
                if '"' in stringtoadd or "'" in stringtoadd or " " in stringtoadd:
                    stringtoadd = repr(stringtoadd)
                h = user32.GetDlgItem(hwnd, ARGV_CMDLINE_DATA)
                oldstr = GetText(h)
                if oldstr and oldstr[-1] != ' ':
                    oldstr = oldstr + ' '
                oldstr = oldstr + stringtoadd
                if oldstr[-1] != ' ':
                    oldstr = oldstr + ' '
                user32.SetWindowTextA(h, oldstr)
                user32.SendMessageA(user32.GetDlgItem(hwnd, ARGV_CMDLINE_DATA), EM_SETSEL, 0, -1)

        return False

    # This next line is needed to prevent gc of the callback
    myDialogProc = DialogProc(DlgProc)
    if IDOK == user32.DialogBoxIndirectParamA(0, resources[id], 0, myDialogProc, 0):
        oldstr = commandLineContents[0]
        tmplist = oldstr.split()
        newlist = []
        while tmplist:
            item = tmplist[0]
            del tmplist[0]
            if item[0] == '"':
                while item[-1] != '"':
                    if not tmplist:
                        raise RuntimeError("Unterminated quoted argument")
                    item = item + ' ' + tmplist[0]
                    del tmplist[0]
                item = item[1:-1]
            if item[0] == "'":
                while item[-1] != "'":
                    if not tmplist:
                        raise RuntimeError("Unterminated quoted argument")
                    item = item + ' ' + tmplist[0]
                    del tmplist[0]
                item = item[1:-1]
            newlist.append(item)
        return newlist
    else:
        raise SystemExit


def test():
    import time

    class empty: pass
    Carbon = empty()
    Carbon.File = empty()
    Carbon.File.FSSpec = None
    Carbon.File.FSRef = None
    MacOS = empty()

    Message("Testing EasyDialogs.")
    optionlist = (('v', 'Verbose'), ('verbose', 'Verbose as long option'),
                ('flags=', 'Valued option'), ('f:', 'Short valued option'))
    commandlist = (('start', 'Start something'), ('stop', 'Stop something'))
    argv = GetArgv(optionlist=optionlist, commandlist=commandlist, addoldfile=0)
    Message("Command line: %s"%' '.join(argv))
    for i in range(len(argv)):
        print(('arg[%d] = %r' % (i, argv[i])))
    ok = AskYesNoCancel("Do you want to proceed?")
    ok = AskYesNoCancel("Do you want to identify?", yes="Identify", no="No")
    if ok > 0:
        s = AskString("Enter your first name", "Joe")
        s2 = AskPassword("Okay %s, tell us your nickname"%s, s, cancel="None")
        if not s2:
            Message("%s has no secret nickname"%s)
        else:
            Message("Hello everybody!!\nThe secret nickname of %s is %s!!!"%(s, s2))
    else:
        s = 'Anonymous'
    rv = AskFileForOpen(message="Gimme a file, %s"%s, wanted=Carbon.File.FSSpec)
    Message("rv: %s"%rv)
    rv = AskFileForSave(wanted=Carbon.File.FSRef, savedFileName="%s.txt"%s)
    Message("rv: %s"%rv) # was: Message("rv.as_pathname: %s"%rv.as_pathname())
    rv = AskFolder()
    Message("Folder name: %s"%rv)
    text = ( "Working Hard...", "Hardly Working..." ,
            "So far, so good!", "Keep on truckin'" )
    bar = ProgressBar("Progress, progress...", 0, label="Ramping up...")
    try:
        if hasattr(MacOS, 'SchedParams'):
            appsw = MacOS.SchedParams(1, 0)
        for i in xrange(20):
            bar.inc()
            time.sleep(0.05)
        bar.set(0,100)
        for i in xrange(100):
            bar.set(i)
            time.sleep(0.05)
            if i % 10 == 0:
                bar.label(text[(i//10) % 4])
        bar.label("Done.")
        time.sleep(1.0)     # give'em a chance to see "Done."
    finally:
        del bar
        if hasattr(MacOS, 'SchedParams'):
            MacOS.SchedParams(*appsw)

if __name__ == '__main__':
    try:
        test()
    except KeyboardInterrupt:
        Message("Operation Canceled.")



from uguugl cimport GLenum
from uguugl cimport GLboolean
from uguugl cimport GLbitfield
from uguugl cimport GLvoid
from uguugl cimport GLbyte
from uguugl cimport GLshort
from uguugl cimport GLint
from uguugl cimport GLclampx
from uguugl cimport GLubyte
from uguugl cimport GLushort
from uguugl cimport GLuint
from uguugl cimport GLsizei
from uguugl cimport GLfloat
from uguugl cimport GLclampf
from uguugl cimport GLdouble
from uguugl cimport GLclampd
from uguugl cimport GLeglClientBufferEXT
from uguugl cimport GLeglImageOES
from uguugl cimport GLchar
from uguugl cimport GLcharARB
from uguugl cimport GLhalfARB
from uguugl cimport GLhalf
from uguugl cimport GLfixed
from uguugl cimport GLintptr
from uguugl cimport GLsizeiptr
from uguugl cimport GLint64
from uguugl cimport GLuint64
from uguugl cimport GLintptrARB
from uguugl cimport GLsizeiptrARB
from uguugl cimport GLint64EXT
from uguugl cimport GLuint64EXT

from libc.stddef cimport ptrdiff_t
from libc.stdint cimport int64_t, uint64_t
from libc.stdio cimport printf
from libc.stdlib cimport calloc, free

from cpython.buffer cimport PyObject_GetBuffer, PyBuffer_Release, PyBUF_CONTIG, PyBUF_CONTIG_RO

cimport uguugl
from uguugl import load

cdef object proxy_return_string(const GLubyte *s):
    """
    This is used for string return values. It returns the return value as
    a python string if it's not NULL, or None if it's null.
    """

    if s == NULL:
        return None

    cdef const char *ss = <const char *> s
    return ss


cdef class ptr:
    """
    This is a class that wraps a generic contiguous Python buffer, and
    allows the retrieval of a pointer to that buffer.
    """

    cdef void *ptr
    cdef Py_buffer view

    def __init__(self, o, ro=True):
        if o is None:
            self.ptr = NULL
            return

        PyObject_GetBuffer(o, &self.view, PyBUF_CONTIG_RO if ro else PyBUF_CONTIG)
        self.ptr = self.view.buf

    def __dealloc__(self):
        if self.ptr:
            PyBuffer_Release(&self.view)
            self.ptr = NULL

cdef ptr get_ptr(o):
    """
    If o is a ptr, return it. Otherwise, convert the buffer into a ptr, and
    return that.
    """

    if isinstance(o, ptr):
        return o
    else:
        return ptr(o)

cdef class Buffer:
    """
    The base class for all buffers.
    """

    cdef Py_ssize_t length
    cdef Py_ssize_t itemsize
    cdef const char *format
    cdef void *data
    cdef int readonly

    cdef setup_buffer(self, Py_ssize_t length, Py_ssize_t itemsize, const char *format, int readonly):
        """
        This is called by a specific buffer's init method to set up various fields, and especially
        allocate the data field.

        `length`
            The number of items in this buffer.
        `itemsize`
            The size of a single item.
        `format`
            The struct-style format code.
        `readonly`
            1 if readonly, 0 if read-write.
        """

        self.length = length
        self.itemsize = itemsize
        self.format = format
        self.readonly = readonly
        self.data = calloc(self.length, self.itemsize)

    def __getbuffer__(self, Py_buffer *buffer, int flags):

        buffer.buf = self.data
        buffer.format = self.format
        buffer.internal = NULL
        buffer.itemsize = self.itemsize
        buffer.len = self.length * self.itemsize
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = self.readonly
        buffer.shape = &self.length
        buffer.strides = &self.itemsize
        buffer.suboffsets = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def __dealloc__(self):
        if self.data:
            free(self.data)
            self.data = NULL

cdef class BytesBuffer(Buffer):

    def __init__(self, length):

        self.setup_buffer(length, 1, "B", 0)

    def get(self):
        return bytes(<char *> self.data)

cdef class BytesListBuffer(Buffer):
    cdef object value

    def __init__(self, value):
        self.value = [ ptr(v) for v in value ]
        self.setup_buffer(len(value), sizeof(const char *), "P", 1)

        cdef int i

        for 0 <= i < self.length:
            (<const char **> self.data)[i] = <const char *> (<ptr> self.value[i]).ptr

cdef class IntBuffer(Buffer):

    def __init__(self, value):

        self.setup_buffer(len(value), sizeof(int), "i", 0)

        cdef int i

        for 0 <= i < self.length:
            (<int*> self.data)[i] = <int> value[i]

    def __getitem__(self, index):
        if index < 0 or index >= self.length:
            raise IndexError("index out of range")

        return (<int*> self.data)[index]

cdef class FloatBuffer(Buffer):

    def __init__(self, value):

        self.setup_buffer(len(value), sizeof(float), "f", 0)

        cdef int i

        for 0 <= i < self.length:
            (<float *> self.data)[i] = <float> value[i]

    def __getitem__(self, index):
        if index < 0 or index >= self.length:
            raise IndexError("index out of range")

        return (<float *> self.data)[index]


from uguugl cimport GLenum
from uguugl cimport GLboolean
from uguugl cimport GLbitfield
from uguugl cimport GLvoid
from uguugl cimport GLbyte
from uguugl cimport GLshort
from uguugl cimport GLint
from uguugl cimport GLclampx
from uguugl cimport GLubyte
from uguugl cimport GLushort
from uguugl cimport GLuint
from uguugl cimport GLsizei
from uguugl cimport GLfloat
from uguugl cimport GLclampf
from uguugl cimport GLdouble
from uguugl cimport GLclampd
from uguugl cimport GLeglClientBufferEXT
from uguugl cimport GLeglImageOES
from uguugl cimport GLchar
from uguugl cimport GLcharARB
from uguugl cimport GLhalfARB
from uguugl cimport GLhalf
from uguugl cimport GLfixed
from uguugl cimport GLintptr
from uguugl cimport GLsizeiptr
from uguugl cimport GLint64
from uguugl cimport GLuint64
from uguugl cimport GLintptrARB
from uguugl cimport GLsizeiptrARB
from uguugl cimport GLint64EXT
from uguugl cimport GLuint64EXT

def glActiveTexture(texture):
    uguugl.glActiveTexture(texture)

def glAttachShader(program, shader):
    uguugl.glAttachShader(program, shader)

def glBindAttribLocation(program, index, name):
    cdef ptr name_ptr = get_ptr(name)
    uguugl.glBindAttribLocation(program, index, <const GLchar *> name_ptr.ptr)

def glBindBuffer(target, buffer):
    uguugl.glBindBuffer(target, buffer)

def glBindFramebuffer(target, framebuffer):
    uguugl.glBindFramebuffer(target, framebuffer)

def glBindRenderbuffer(target, renderbuffer):
    uguugl.glBindRenderbuffer(target, renderbuffer)

def glBindTexture(target, texture):
    uguugl.glBindTexture(target, texture)

def glBlendColor(red, green, blue, alpha):
    uguugl.glBlendColor(red, green, blue, alpha)

def glBlendEquation(mode):
    uguugl.glBlendEquation(mode)

def glBlendEquationSeparate(modeRGB, modeAlpha):
    uguugl.glBlendEquationSeparate(modeRGB, modeAlpha)

def glBlendFunc(sfactor, dfactor):
    uguugl.glBlendFunc(sfactor, dfactor)

def glBlendFuncSeparate(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha):
    uguugl.glBlendFuncSeparate(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha)

def glBufferData(target, size, data, usage):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glBufferData(target, size, <const void *> data_ptr.ptr, usage)

def glBufferSubData(target, offset, size, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glBufferSubData(target, offset, size, <const void *> data_ptr.ptr)

def glCheckFramebufferStatus(target):
    return uguugl.glCheckFramebufferStatus(target)

def glClear(mask):
    uguugl.glClear(mask)

def glClearColor(red, green, blue, alpha):
    uguugl.glClearColor(red, green, blue, alpha)

def glClearStencil(s):
    uguugl.glClearStencil(s)

def glColorMask(red, green, blue, alpha):
    uguugl.glColorMask(red, green, blue, alpha)

def glCompileShader(shader):
    uguugl.glCompileShader(shader)

def glCompressedTexImage2D(target, level, internalformat, width, height, border, imageSize, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glCompressedTexImage2D(target, level, internalformat, width, height, border, imageSize, <const void *> data_ptr.ptr)

def glCompressedTexSubImage2D(target, level, xoffset, yoffset, width, height, format, imageSize, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glCompressedTexSubImage2D(target, level, xoffset, yoffset, width, height, format, imageSize, <const void *> data_ptr.ptr)

def glCopyTexImage2D(target, level, internalformat, x, y, width, height, border):
    uguugl.glCopyTexImage2D(target, level, internalformat, x, y, width, height, border)

def glCopyTexSubImage2D(target, level, xoffset, yoffset, x, y, width, height):
    uguugl.glCopyTexSubImage2D(target, level, xoffset, yoffset, x, y, width, height)

def glCreateProgram():
    return uguugl.glCreateProgram()

def glCreateShader(type):
    return uguugl.glCreateShader(type)

def glCullFace(mode):
    uguugl.glCullFace(mode)

def glDeleteBuffers(n, buffers):
    cdef ptr buffers_ptr = get_ptr(buffers)
    uguugl.glDeleteBuffers(n, <const GLuint *> buffers_ptr.ptr)

def glDeleteFramebuffers(n, framebuffers):
    cdef ptr framebuffers_ptr = get_ptr(framebuffers)
    uguugl.glDeleteFramebuffers(n, <const GLuint *> framebuffers_ptr.ptr)

def glDeleteProgram(program):
    uguugl.glDeleteProgram(program)

def glDeleteRenderbuffers(n, renderbuffers):
    cdef ptr renderbuffers_ptr = get_ptr(renderbuffers)
    uguugl.glDeleteRenderbuffers(n, <const GLuint *> renderbuffers_ptr.ptr)

def glDeleteShader(shader):
    uguugl.glDeleteShader(shader)

def glDeleteTextures(n, textures):
    cdef ptr textures_ptr = get_ptr(textures)
    uguugl.glDeleteTextures(n, <const GLuint *> textures_ptr.ptr)

def glDepthFunc(func):
    uguugl.glDepthFunc(func)

def glDepthMask(flag):
    uguugl.glDepthMask(flag)

def glDetachShader(program, shader):
    uguugl.glDetachShader(program, shader)

def glDisable(cap):
    uguugl.glDisable(cap)

def glDisableVertexAttribArray(index):
    uguugl.glDisableVertexAttribArray(index)

def glDrawArrays(mode, first, count):
    uguugl.glDrawArrays(mode, first, count)

def glDrawElements(mode, count, type, indices):
    cdef ptr indices_ptr = get_ptr(indices)
    uguugl.glDrawElements(mode, count, type, <const void *> indices_ptr.ptr)

def glEnable(cap):
    uguugl.glEnable(cap)

def glEnableVertexAttribArray(index):
    uguugl.glEnableVertexAttribArray(index)

def glFinish():
    uguugl.glFinish()

def glFlush():
    uguugl.glFlush()

def glFramebufferRenderbuffer(target, attachment, renderbuffertarget, renderbuffer):
    uguugl.glFramebufferRenderbuffer(target, attachment, renderbuffertarget, renderbuffer)

def glFramebufferTexture2D(target, attachment, textarget, texture, level):
    uguugl.glFramebufferTexture2D(target, attachment, textarget, texture, level)

def glFrontFace(mode):
    uguugl.glFrontFace(mode)

def glGenBuffers(n, buffers):
    cdef ptr buffers_ptr = get_ptr(buffers)
    uguugl.glGenBuffers(n, <GLuint *> buffers_ptr.ptr)

def glGenFramebuffers(n, framebuffers):
    cdef ptr framebuffers_ptr = get_ptr(framebuffers)
    uguugl.glGenFramebuffers(n, <GLuint *> framebuffers_ptr.ptr)

def glGenRenderbuffers(n, renderbuffers):
    cdef ptr renderbuffers_ptr = get_ptr(renderbuffers)
    uguugl.glGenRenderbuffers(n, <GLuint *> renderbuffers_ptr.ptr)

def glGenTextures(n, textures):
    cdef ptr textures_ptr = get_ptr(textures)
    uguugl.glGenTextures(n, <GLuint *> textures_ptr.ptr)

def glGenerateMipmap(target):
    uguugl.glGenerateMipmap(target)

def glGetActiveAttrib(program, index, bufSize, length, size, type, name):
    cdef ptr length_ptr = get_ptr(length)
    cdef ptr size_ptr = get_ptr(size)
    cdef ptr type_ptr = get_ptr(type)
    cdef ptr name_ptr = get_ptr(name)
    uguugl.glGetActiveAttrib(program, index, bufSize, <GLsizei *> length_ptr.ptr, <GLint *> size_ptr.ptr, <GLenum *> type_ptr.ptr, <GLchar *> name_ptr.ptr)

def glGetActiveUniform(program, index, bufSize, length, size, type, name):
    cdef ptr length_ptr = get_ptr(length)
    cdef ptr size_ptr = get_ptr(size)
    cdef ptr type_ptr = get_ptr(type)
    cdef ptr name_ptr = get_ptr(name)
    uguugl.glGetActiveUniform(program, index, bufSize, <GLsizei *> length_ptr.ptr, <GLint *> size_ptr.ptr, <GLenum *> type_ptr.ptr, <GLchar *> name_ptr.ptr)

def glGetAttachedShaders(program, maxCount, count, shaders):
    cdef ptr count_ptr = get_ptr(count)
    cdef ptr shaders_ptr = get_ptr(shaders)
    uguugl.glGetAttachedShaders(program, maxCount, <GLsizei *> count_ptr.ptr, <GLuint *> shaders_ptr.ptr)

def glGetAttribLocation(program, name):
    cdef ptr name_ptr = get_ptr(name)
    return uguugl.glGetAttribLocation(program, <const GLchar *> name_ptr.ptr)

def glGetBooleanv(pname, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glGetBooleanv(pname, <GLboolean *> data_ptr.ptr)

def glGetBufferParameteriv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetBufferParameteriv(target, pname, <GLint *> params_ptr.ptr)

def glGetError():
    return uguugl.glGetError()

def glGetFloatv(pname, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glGetFloatv(pname, <GLfloat *> data_ptr.ptr)

def glGetFramebufferAttachmentParameteriv(target, attachment, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetFramebufferAttachmentParameteriv(target, attachment, pname, <GLint *> params_ptr.ptr)

def glGetIntegerv(pname, data):
    cdef ptr data_ptr = get_ptr(data)
    uguugl.glGetIntegerv(pname, <GLint *> data_ptr.ptr)

def glGetProgramInfoLog(program, bufSize, length, infoLog):
    cdef ptr length_ptr = get_ptr(length)
    cdef ptr infoLog_ptr = get_ptr(infoLog)
    uguugl.glGetProgramInfoLog(program, bufSize, <GLsizei *> length_ptr.ptr, <GLchar *> infoLog_ptr.ptr)

def glGetProgramiv(program, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetProgramiv(program, pname, <GLint *> params_ptr.ptr)

def glGetRenderbufferParameteriv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetRenderbufferParameteriv(target, pname, <GLint *> params_ptr.ptr)

def glGetShaderInfoLog(shader, bufSize, length, infoLog):
    cdef ptr length_ptr = get_ptr(length)
    cdef ptr infoLog_ptr = get_ptr(infoLog)
    uguugl.glGetShaderInfoLog(shader, bufSize, <GLsizei *> length_ptr.ptr, <GLchar *> infoLog_ptr.ptr)

def glGetShaderSource(shader, bufSize, length, source):
    cdef ptr length_ptr = get_ptr(length)
    cdef ptr source_ptr = get_ptr(source)
    uguugl.glGetShaderSource(shader, bufSize, <GLsizei *> length_ptr.ptr, <GLchar *> source_ptr.ptr)

def glGetShaderiv(shader, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetShaderiv(shader, pname, <GLint *> params_ptr.ptr)

def glGetString(name):
    return proxy_return_string(uguugl.glGetString(name))

def glGetTexParameterfv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetTexParameterfv(target, pname, <GLfloat *> params_ptr.ptr)

def glGetTexParameteriv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetTexParameteriv(target, pname, <GLint *> params_ptr.ptr)

def glGetUniformLocation(program, name):
    cdef ptr name_ptr = get_ptr(name)
    return uguugl.glGetUniformLocation(program, <const GLchar *> name_ptr.ptr)

def glGetUniformfv(program, location, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetUniformfv(program, location, <GLfloat *> params_ptr.ptr)

def glGetUniformiv(program, location, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetUniformiv(program, location, <GLint *> params_ptr.ptr)

def glGetVertexAttribPointerv(index, pname, pointer):
    cdef ptr pointer_ptr = get_ptr(pointer)
    uguugl.glGetVertexAttribPointerv(index, pname, <void **> pointer_ptr.ptr)

def glGetVertexAttribfv(index, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetVertexAttribfv(index, pname, <GLfloat *> params_ptr.ptr)

def glGetVertexAttribiv(index, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glGetVertexAttribiv(index, pname, <GLint *> params_ptr.ptr)

def glHint(target, mode):
    uguugl.glHint(target, mode)

def glIsBuffer(buffer):
    return uguugl.glIsBuffer(buffer)

def glIsEnabled(cap):
    return uguugl.glIsEnabled(cap)

def glIsFramebuffer(framebuffer):
    return uguugl.glIsFramebuffer(framebuffer)

def glIsProgram(program):
    return uguugl.glIsProgram(program)

def glIsRenderbuffer(renderbuffer):
    return uguugl.glIsRenderbuffer(renderbuffer)

def glIsShader(shader):
    return uguugl.glIsShader(shader)

def glIsTexture(texture):
    return uguugl.glIsTexture(texture)

def glLineWidth(width):
    uguugl.glLineWidth(width)

def glLinkProgram(program):
    uguugl.glLinkProgram(program)

def glPixelStorei(pname, param):
    uguugl.glPixelStorei(pname, param)

def glPolygonOffset(factor, units):
    uguugl.glPolygonOffset(factor, units)

def glReadPixels(x, y, width, height, format, type, pixels):
    cdef ptr pixels_ptr = get_ptr(pixels)
    uguugl.glReadPixels(x, y, width, height, format, type, <void *> pixels_ptr.ptr)

def glRenderbufferStorage(target, internalformat, width, height):
    uguugl.glRenderbufferStorage(target, internalformat, width, height)

def glSampleCoverage(value, invert):
    uguugl.glSampleCoverage(value, invert)

def glScissor(x, y, width, height):
    uguugl.glScissor(x, y, width, height)

def glShaderSource(shader, count, string, length):
    cdef ptr string_ptr = get_ptr(string)
    cdef ptr length_ptr = get_ptr(length)
    uguugl.glShaderSource(shader, count, <const GLchar *const*> string_ptr.ptr, <const GLint *> length_ptr.ptr)

def glStencilFunc(func, ref, mask):
    uguugl.glStencilFunc(func, ref, mask)

def glStencilFuncSeparate(face, func, ref, mask):
    uguugl.glStencilFuncSeparate(face, func, ref, mask)

def glStencilMask(mask):
    uguugl.glStencilMask(mask)

def glStencilMaskSeparate(face, mask):
    uguugl.glStencilMaskSeparate(face, mask)

def glStencilOp(fail, zfail, zpass):
    uguugl.glStencilOp(fail, zfail, zpass)

def glStencilOpSeparate(face, sfail, dpfail, dppass):
    uguugl.glStencilOpSeparate(face, sfail, dpfail, dppass)

def glTexImage2D(target, level, internalformat, width, height, border, format, type, pixels):
    cdef ptr pixels_ptr = get_ptr(pixels)
    uguugl.glTexImage2D(target, level, internalformat, width, height, border, format, type, <const void *> pixels_ptr.ptr)

def glTexParameterf(target, pname, param):
    uguugl.glTexParameterf(target, pname, param)

def glTexParameterfv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glTexParameterfv(target, pname, <const GLfloat *> params_ptr.ptr)

def glTexParameteri(target, pname, param):
    uguugl.glTexParameteri(target, pname, param)

def glTexParameteriv(target, pname, params):
    cdef ptr params_ptr = get_ptr(params)
    uguugl.glTexParameteriv(target, pname, <const GLint *> params_ptr.ptr)

def glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels):
    cdef ptr pixels_ptr = get_ptr(pixels)
    uguugl.glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, <const void *> pixels_ptr.ptr)

def glUniform1f(location, v0):
    uguugl.glUniform1f(location, v0)

def glUniform1fv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform1fv(location, count, <const GLfloat *> value_ptr.ptr)

def glUniform1i(location, v0):
    uguugl.glUniform1i(location, v0)

def glUniform1iv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform1iv(location, count, <const GLint *> value_ptr.ptr)

def glUniform2f(location, v0, v1):
    uguugl.glUniform2f(location, v0, v1)

def glUniform2fv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform2fv(location, count, <const GLfloat *> value_ptr.ptr)

def glUniform2i(location, v0, v1):
    uguugl.glUniform2i(location, v0, v1)

def glUniform2iv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform2iv(location, count, <const GLint *> value_ptr.ptr)

def glUniform3f(location, v0, v1, v2):
    uguugl.glUniform3f(location, v0, v1, v2)

def glUniform3fv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform3fv(location, count, <const GLfloat *> value_ptr.ptr)

def glUniform3i(location, v0, v1, v2):
    uguugl.glUniform3i(location, v0, v1, v2)

def glUniform3iv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform3iv(location, count, <const GLint *> value_ptr.ptr)

def glUniform4f(location, v0, v1, v2, v3):
    uguugl.glUniform4f(location, v0, v1, v2, v3)

def glUniform4fv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform4fv(location, count, <const GLfloat *> value_ptr.ptr)

def glUniform4i(location, v0, v1, v2, v3):
    uguugl.glUniform4i(location, v0, v1, v2, v3)

def glUniform4iv(location, count, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniform4iv(location, count, <const GLint *> value_ptr.ptr)

def glUniformMatrix2fv(location, count, transpose, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniformMatrix2fv(location, count, transpose, <const GLfloat *> value_ptr.ptr)

def glUniformMatrix3fv(location, count, transpose, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniformMatrix3fv(location, count, transpose, <const GLfloat *> value_ptr.ptr)

def glUniformMatrix4fv(location, count, transpose, value):
    cdef ptr value_ptr = get_ptr(value)
    uguugl.glUniformMatrix4fv(location, count, transpose, <const GLfloat *> value_ptr.ptr)

def glUseProgram(program):
    uguugl.glUseProgram(program)

def glValidateProgram(program):
    uguugl.glValidateProgram(program)

def glVertexAttrib1f(index, x):
    uguugl.glVertexAttrib1f(index, x)

def glVertexAttrib1fv(index, v):
    cdef ptr v_ptr = get_ptr(v)
    uguugl.glVertexAttrib1fv(index, <const GLfloat *> v_ptr.ptr)

def glVertexAttrib2f(index, x, y):
    uguugl.glVertexAttrib2f(index, x, y)

def glVertexAttrib2fv(index, v):
    cdef ptr v_ptr = get_ptr(v)
    uguugl.glVertexAttrib2fv(index, <const GLfloat *> v_ptr.ptr)

def glVertexAttrib3f(index, x, y, z):
    uguugl.glVertexAttrib3f(index, x, y, z)

def glVertexAttrib3fv(index, v):
    cdef ptr v_ptr = get_ptr(v)
    uguugl.glVertexAttrib3fv(index, <const GLfloat *> v_ptr.ptr)

def glVertexAttrib4f(index, x, y, z, w):
    uguugl.glVertexAttrib4f(index, x, y, z, w)

def glVertexAttrib4fv(index, v):
    cdef ptr v_ptr = get_ptr(v)
    uguugl.glVertexAttrib4fv(index, <const GLfloat *> v_ptr.ptr)

def glVertexAttribPointer(index, size, type, normalized, stride, pointer):
    cdef ptr pointer_ptr = get_ptr(pointer)
    uguugl.glVertexAttribPointer(index, size, type, normalized, stride, <const void *> pointer_ptr.ptr)

def glViewport(x, y, width, height):
    uguugl.glViewport(x, y, width, height)

GL_FALSE = uguugl.GL_FALSE
GL_NONE = uguugl.GL_NONE
GL_NO_ERROR = uguugl.GL_NO_ERROR
GL_POINTS = uguugl.GL_POINTS
GL_ZERO = uguugl.GL_ZERO
GL_LINES = uguugl.GL_LINES
GL_ONE = uguugl.GL_ONE
GL_TRUE = uguugl.GL_TRUE
GL_LINE_LOOP = uguugl.GL_LINE_LOOP
GL_LINE_STRIP = uguugl.GL_LINE_STRIP
GL_TRIANGLES = uguugl.GL_TRIANGLES
GL_TRIANGLE_STRIP = uguugl.GL_TRIANGLE_STRIP
GL_TRIANGLE_FAN = uguugl.GL_TRIANGLE_FAN
GL_DEPTH_BUFFER_BIT = uguugl.GL_DEPTH_BUFFER_BIT
GL_NEVER = uguugl.GL_NEVER
GL_LESS = uguugl.GL_LESS
GL_EQUAL = uguugl.GL_EQUAL
GL_LEQUAL = uguugl.GL_LEQUAL
GL_GREATER = uguugl.GL_GREATER
GL_NOTEQUAL = uguugl.GL_NOTEQUAL
GL_GEQUAL = uguugl.GL_GEQUAL
GL_ALWAYS = uguugl.GL_ALWAYS
GL_SRC_COLOR = uguugl.GL_SRC_COLOR
GL_ONE_MINUS_SRC_COLOR = uguugl.GL_ONE_MINUS_SRC_COLOR
GL_SRC_ALPHA = uguugl.GL_SRC_ALPHA
GL_ONE_MINUS_SRC_ALPHA = uguugl.GL_ONE_MINUS_SRC_ALPHA
GL_DST_ALPHA = uguugl.GL_DST_ALPHA
GL_ONE_MINUS_DST_ALPHA = uguugl.GL_ONE_MINUS_DST_ALPHA
GL_DST_COLOR = uguugl.GL_DST_COLOR
GL_ONE_MINUS_DST_COLOR = uguugl.GL_ONE_MINUS_DST_COLOR
GL_SRC_ALPHA_SATURATE = uguugl.GL_SRC_ALPHA_SATURATE
GL_STENCIL_BUFFER_BIT = uguugl.GL_STENCIL_BUFFER_BIT
GL_FRONT = uguugl.GL_FRONT
GL_BACK = uguugl.GL_BACK
GL_FRONT_AND_BACK = uguugl.GL_FRONT_AND_BACK
GL_INVALID_ENUM = uguugl.GL_INVALID_ENUM
GL_INVALID_VALUE = uguugl.GL_INVALID_VALUE
GL_INVALID_OPERATION = uguugl.GL_INVALID_OPERATION
GL_OUT_OF_MEMORY = uguugl.GL_OUT_OF_MEMORY
GL_INVALID_FRAMEBUFFER_OPERATION = uguugl.GL_INVALID_FRAMEBUFFER_OPERATION
GL_CW = uguugl.GL_CW
GL_CCW = uguugl.GL_CCW
GL_LINE_WIDTH = uguugl.GL_LINE_WIDTH
GL_CULL_FACE = uguugl.GL_CULL_FACE
GL_CULL_FACE_MODE = uguugl.GL_CULL_FACE_MODE
GL_FRONT_FACE = uguugl.GL_FRONT_FACE
GL_DEPTH_RANGE = uguugl.GL_DEPTH_RANGE
GL_DEPTH_TEST = uguugl.GL_DEPTH_TEST
GL_DEPTH_WRITEMASK = uguugl.GL_DEPTH_WRITEMASK
GL_DEPTH_CLEAR_VALUE = uguugl.GL_DEPTH_CLEAR_VALUE
GL_DEPTH_FUNC = uguugl.GL_DEPTH_FUNC
GL_STENCIL_TEST = uguugl.GL_STENCIL_TEST
GL_STENCIL_CLEAR_VALUE = uguugl.GL_STENCIL_CLEAR_VALUE
GL_STENCIL_FUNC = uguugl.GL_STENCIL_FUNC
GL_STENCIL_VALUE_MASK = uguugl.GL_STENCIL_VALUE_MASK
GL_STENCIL_FAIL = uguugl.GL_STENCIL_FAIL
GL_STENCIL_PASS_DEPTH_FAIL = uguugl.GL_STENCIL_PASS_DEPTH_FAIL
GL_STENCIL_PASS_DEPTH_PASS = uguugl.GL_STENCIL_PASS_DEPTH_PASS
GL_STENCIL_REF = uguugl.GL_STENCIL_REF
GL_STENCIL_WRITEMASK = uguugl.GL_STENCIL_WRITEMASK
GL_VIEWPORT = uguugl.GL_VIEWPORT
GL_DITHER = uguugl.GL_DITHER
GL_BLEND = uguugl.GL_BLEND
GL_SCISSOR_BOX = uguugl.GL_SCISSOR_BOX
GL_SCISSOR_TEST = uguugl.GL_SCISSOR_TEST
GL_COLOR_CLEAR_VALUE = uguugl.GL_COLOR_CLEAR_VALUE
GL_COLOR_WRITEMASK = uguugl.GL_COLOR_WRITEMASK
GL_UNPACK_ALIGNMENT = uguugl.GL_UNPACK_ALIGNMENT
GL_PACK_ALIGNMENT = uguugl.GL_PACK_ALIGNMENT
GL_MAX_TEXTURE_SIZE = uguugl.GL_MAX_TEXTURE_SIZE
GL_MAX_VIEWPORT_DIMS = uguugl.GL_MAX_VIEWPORT_DIMS
GL_SUBPIXEL_BITS = uguugl.GL_SUBPIXEL_BITS
GL_RED_BITS = uguugl.GL_RED_BITS
GL_GREEN_BITS = uguugl.GL_GREEN_BITS
GL_BLUE_BITS = uguugl.GL_BLUE_BITS
GL_ALPHA_BITS = uguugl.GL_ALPHA_BITS
GL_DEPTH_BITS = uguugl.GL_DEPTH_BITS
GL_STENCIL_BITS = uguugl.GL_STENCIL_BITS
GL_TEXTURE_2D = uguugl.GL_TEXTURE_2D
GL_DONT_CARE = uguugl.GL_DONT_CARE
GL_FASTEST = uguugl.GL_FASTEST
GL_NICEST = uguugl.GL_NICEST
GL_BYTE = uguugl.GL_BYTE
GL_UNSIGNED_BYTE = uguugl.GL_UNSIGNED_BYTE
GL_SHORT = uguugl.GL_SHORT
GL_UNSIGNED_SHORT = uguugl.GL_UNSIGNED_SHORT
GL_INT = uguugl.GL_INT
GL_UNSIGNED_INT = uguugl.GL_UNSIGNED_INT
GL_FLOAT = uguugl.GL_FLOAT
GL_INVERT = uguugl.GL_INVERT
GL_TEXTURE = uguugl.GL_TEXTURE
GL_DEPTH_COMPONENT = uguugl.GL_DEPTH_COMPONENT
GL_ALPHA = uguugl.GL_ALPHA
GL_RGB = uguugl.GL_RGB
GL_RGBA = uguugl.GL_RGBA
GL_LUMINANCE = uguugl.GL_LUMINANCE
GL_LUMINANCE_ALPHA = uguugl.GL_LUMINANCE_ALPHA
GL_KEEP = uguugl.GL_KEEP
GL_REPLACE = uguugl.GL_REPLACE
GL_INCR = uguugl.GL_INCR
GL_DECR = uguugl.GL_DECR
GL_VENDOR = uguugl.GL_VENDOR
GL_RENDERER = uguugl.GL_RENDERER
GL_VERSION = uguugl.GL_VERSION
GL_EXTENSIONS = uguugl.GL_EXTENSIONS
GL_NEAREST = uguugl.GL_NEAREST
GL_LINEAR = uguugl.GL_LINEAR
GL_NEAREST_MIPMAP_NEAREST = uguugl.GL_NEAREST_MIPMAP_NEAREST
GL_LINEAR_MIPMAP_NEAREST = uguugl.GL_LINEAR_MIPMAP_NEAREST
GL_NEAREST_MIPMAP_LINEAR = uguugl.GL_NEAREST_MIPMAP_LINEAR
GL_LINEAR_MIPMAP_LINEAR = uguugl.GL_LINEAR_MIPMAP_LINEAR
GL_TEXTURE_MAG_FILTER = uguugl.GL_TEXTURE_MAG_FILTER
GL_TEXTURE_MIN_FILTER = uguugl.GL_TEXTURE_MIN_FILTER
GL_TEXTURE_WRAP_S = uguugl.GL_TEXTURE_WRAP_S
GL_TEXTURE_WRAP_T = uguugl.GL_TEXTURE_WRAP_T
GL_REPEAT = uguugl.GL_REPEAT
GL_POLYGON_OFFSET_UNITS = uguugl.GL_POLYGON_OFFSET_UNITS
GL_COLOR_BUFFER_BIT = uguugl.GL_COLOR_BUFFER_BIT
GL_CONSTANT_COLOR = uguugl.GL_CONSTANT_COLOR
GL_ONE_MINUS_CONSTANT_COLOR = uguugl.GL_ONE_MINUS_CONSTANT_COLOR
GL_CONSTANT_ALPHA = uguugl.GL_CONSTANT_ALPHA
GL_ONE_MINUS_CONSTANT_ALPHA = uguugl.GL_ONE_MINUS_CONSTANT_ALPHA
GL_BLEND_COLOR = uguugl.GL_BLEND_COLOR
GL_FUNC_ADD = uguugl.GL_FUNC_ADD
GL_BLEND_EQUATION = uguugl.GL_BLEND_EQUATION
GL_BLEND_EQUATION_RGB = uguugl.GL_BLEND_EQUATION_RGB
GL_FUNC_SUBTRACT = uguugl.GL_FUNC_SUBTRACT
GL_FUNC_REVERSE_SUBTRACT = uguugl.GL_FUNC_REVERSE_SUBTRACT
GL_UNSIGNED_SHORT_4_4_4_4 = uguugl.GL_UNSIGNED_SHORT_4_4_4_4
GL_UNSIGNED_SHORT_5_5_5_1 = uguugl.GL_UNSIGNED_SHORT_5_5_5_1
GL_POLYGON_OFFSET_FILL = uguugl.GL_POLYGON_OFFSET_FILL
GL_POLYGON_OFFSET_FACTOR = uguugl.GL_POLYGON_OFFSET_FACTOR
GL_RGBA4 = uguugl.GL_RGBA4
GL_RGB5_A1 = uguugl.GL_RGB5_A1
GL_TEXTURE_BINDING_2D = uguugl.GL_TEXTURE_BINDING_2D
GL_SAMPLE_ALPHA_TO_COVERAGE = uguugl.GL_SAMPLE_ALPHA_TO_COVERAGE
GL_SAMPLE_COVERAGE = uguugl.GL_SAMPLE_COVERAGE
GL_SAMPLE_BUFFERS = uguugl.GL_SAMPLE_BUFFERS
GL_SAMPLES = uguugl.GL_SAMPLES
GL_SAMPLE_COVERAGE_VALUE = uguugl.GL_SAMPLE_COVERAGE_VALUE
GL_SAMPLE_COVERAGE_INVERT = uguugl.GL_SAMPLE_COVERAGE_INVERT
GL_BLEND_DST_RGB = uguugl.GL_BLEND_DST_RGB
GL_BLEND_SRC_RGB = uguugl.GL_BLEND_SRC_RGB
GL_BLEND_DST_ALPHA = uguugl.GL_BLEND_DST_ALPHA
GL_BLEND_SRC_ALPHA = uguugl.GL_BLEND_SRC_ALPHA
GL_CLAMP_TO_EDGE = uguugl.GL_CLAMP_TO_EDGE
GL_GENERATE_MIPMAP_HINT = uguugl.GL_GENERATE_MIPMAP_HINT
GL_DEPTH_COMPONENT16 = uguugl.GL_DEPTH_COMPONENT16
GL_UNSIGNED_SHORT_5_6_5 = uguugl.GL_UNSIGNED_SHORT_5_6_5
GL_MIRRORED_REPEAT = uguugl.GL_MIRRORED_REPEAT
GL_ALIASED_POINT_SIZE_RANGE = uguugl.GL_ALIASED_POINT_SIZE_RANGE
GL_ALIASED_LINE_WIDTH_RANGE = uguugl.GL_ALIASED_LINE_WIDTH_RANGE
GL_TEXTURE0 = uguugl.GL_TEXTURE0
GL_TEXTURE1 = uguugl.GL_TEXTURE1
GL_TEXTURE2 = uguugl.GL_TEXTURE2
GL_TEXTURE3 = uguugl.GL_TEXTURE3
GL_TEXTURE4 = uguugl.GL_TEXTURE4
GL_TEXTURE5 = uguugl.GL_TEXTURE5
GL_TEXTURE6 = uguugl.GL_TEXTURE6
GL_TEXTURE7 = uguugl.GL_TEXTURE7
GL_TEXTURE8 = uguugl.GL_TEXTURE8
GL_TEXTURE9 = uguugl.GL_TEXTURE9
GL_TEXTURE10 = uguugl.GL_TEXTURE10
GL_TEXTURE11 = uguugl.GL_TEXTURE11
GL_TEXTURE12 = uguugl.GL_TEXTURE12
GL_TEXTURE13 = uguugl.GL_TEXTURE13
GL_TEXTURE14 = uguugl.GL_TEXTURE14
GL_TEXTURE15 = uguugl.GL_TEXTURE15
GL_TEXTURE16 = uguugl.GL_TEXTURE16
GL_TEXTURE17 = uguugl.GL_TEXTURE17
GL_TEXTURE18 = uguugl.GL_TEXTURE18
GL_TEXTURE19 = uguugl.GL_TEXTURE19
GL_TEXTURE20 = uguugl.GL_TEXTURE20
GL_TEXTURE21 = uguugl.GL_TEXTURE21
GL_TEXTURE22 = uguugl.GL_TEXTURE22
GL_TEXTURE23 = uguugl.GL_TEXTURE23
GL_TEXTURE24 = uguugl.GL_TEXTURE24
GL_TEXTURE25 = uguugl.GL_TEXTURE25
GL_TEXTURE26 = uguugl.GL_TEXTURE26
GL_TEXTURE27 = uguugl.GL_TEXTURE27
GL_TEXTURE28 = uguugl.GL_TEXTURE28
GL_TEXTURE29 = uguugl.GL_TEXTURE29
GL_TEXTURE30 = uguugl.GL_TEXTURE30
GL_TEXTURE31 = uguugl.GL_TEXTURE31
GL_ACTIVE_TEXTURE = uguugl.GL_ACTIVE_TEXTURE
GL_MAX_RENDERBUFFER_SIZE = uguugl.GL_MAX_RENDERBUFFER_SIZE
GL_INCR_WRAP = uguugl.GL_INCR_WRAP
GL_DECR_WRAP = uguugl.GL_DECR_WRAP
GL_TEXTURE_CUBE_MAP = uguugl.GL_TEXTURE_CUBE_MAP
GL_TEXTURE_BINDING_CUBE_MAP = uguugl.GL_TEXTURE_BINDING_CUBE_MAP
GL_TEXTURE_CUBE_MAP_POSITIVE_X = uguugl.GL_TEXTURE_CUBE_MAP_POSITIVE_X
GL_TEXTURE_CUBE_MAP_NEGATIVE_X = uguugl.GL_TEXTURE_CUBE_MAP_NEGATIVE_X
GL_TEXTURE_CUBE_MAP_POSITIVE_Y = uguugl.GL_TEXTURE_CUBE_MAP_POSITIVE_Y
GL_TEXTURE_CUBE_MAP_NEGATIVE_Y = uguugl.GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
GL_TEXTURE_CUBE_MAP_POSITIVE_Z = uguugl.GL_TEXTURE_CUBE_MAP_POSITIVE_Z
GL_TEXTURE_CUBE_MAP_NEGATIVE_Z = uguugl.GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
GL_MAX_CUBE_MAP_TEXTURE_SIZE = uguugl.GL_MAX_CUBE_MAP_TEXTURE_SIZE
GL_VERTEX_ATTRIB_ARRAY_ENABLED = uguugl.GL_VERTEX_ATTRIB_ARRAY_ENABLED
GL_VERTEX_ATTRIB_ARRAY_SIZE = uguugl.GL_VERTEX_ATTRIB_ARRAY_SIZE
GL_VERTEX_ATTRIB_ARRAY_STRIDE = uguugl.GL_VERTEX_ATTRIB_ARRAY_STRIDE
GL_VERTEX_ATTRIB_ARRAY_TYPE = uguugl.GL_VERTEX_ATTRIB_ARRAY_TYPE
GL_CURRENT_VERTEX_ATTRIB = uguugl.GL_CURRENT_VERTEX_ATTRIB
GL_VERTEX_ATTRIB_ARRAY_POINTER = uguugl.GL_VERTEX_ATTRIB_ARRAY_POINTER
GL_NUM_COMPRESSED_TEXTURE_FORMATS = uguugl.GL_NUM_COMPRESSED_TEXTURE_FORMATS
GL_COMPRESSED_TEXTURE_FORMATS = uguugl.GL_COMPRESSED_TEXTURE_FORMATS
GL_BUFFER_SIZE = uguugl.GL_BUFFER_SIZE
GL_BUFFER_USAGE = uguugl.GL_BUFFER_USAGE
GL_STENCIL_BACK_FUNC = uguugl.GL_STENCIL_BACK_FUNC
GL_STENCIL_BACK_FAIL = uguugl.GL_STENCIL_BACK_FAIL
GL_STENCIL_BACK_PASS_DEPTH_FAIL = uguugl.GL_STENCIL_BACK_PASS_DEPTH_FAIL
GL_STENCIL_BACK_PASS_DEPTH_PASS = uguugl.GL_STENCIL_BACK_PASS_DEPTH_PASS
GL_BLEND_EQUATION_ALPHA = uguugl.GL_BLEND_EQUATION_ALPHA
GL_MAX_VERTEX_ATTRIBS = uguugl.GL_MAX_VERTEX_ATTRIBS
GL_VERTEX_ATTRIB_ARRAY_NORMALIZED = uguugl.GL_VERTEX_ATTRIB_ARRAY_NORMALIZED
GL_MAX_TEXTURE_IMAGE_UNITS = uguugl.GL_MAX_TEXTURE_IMAGE_UNITS
GL_ARRAY_BUFFER = uguugl.GL_ARRAY_BUFFER
GL_ELEMENT_ARRAY_BUFFER = uguugl.GL_ELEMENT_ARRAY_BUFFER
GL_ARRAY_BUFFER_BINDING = uguugl.GL_ARRAY_BUFFER_BINDING
GL_ELEMENT_ARRAY_BUFFER_BINDING = uguugl.GL_ELEMENT_ARRAY_BUFFER_BINDING
GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING = uguugl.GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING
GL_STREAM_DRAW = uguugl.GL_STREAM_DRAW
GL_STATIC_DRAW = uguugl.GL_STATIC_DRAW
GL_DYNAMIC_DRAW = uguugl.GL_DYNAMIC_DRAW
GL_FRAGMENT_SHADER = uguugl.GL_FRAGMENT_SHADER
GL_VERTEX_SHADER = uguugl.GL_VERTEX_SHADER
GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS = uguugl.GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS
GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS = uguugl.GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS
GL_SHADER_TYPE = uguugl.GL_SHADER_TYPE
GL_FLOAT_VEC2 = uguugl.GL_FLOAT_VEC2
GL_FLOAT_VEC3 = uguugl.GL_FLOAT_VEC3
GL_FLOAT_VEC4 = uguugl.GL_FLOAT_VEC4
GL_INT_VEC2 = uguugl.GL_INT_VEC2
GL_INT_VEC3 = uguugl.GL_INT_VEC3
GL_INT_VEC4 = uguugl.GL_INT_VEC4
GL_BOOL = uguugl.GL_BOOL
GL_BOOL_VEC2 = uguugl.GL_BOOL_VEC2
GL_BOOL_VEC3 = uguugl.GL_BOOL_VEC3
GL_BOOL_VEC4 = uguugl.GL_BOOL_VEC4
GL_FLOAT_MAT2 = uguugl.GL_FLOAT_MAT2
GL_FLOAT_MAT3 = uguugl.GL_FLOAT_MAT3
GL_FLOAT_MAT4 = uguugl.GL_FLOAT_MAT4
GL_SAMPLER_2D = uguugl.GL_SAMPLER_2D
GL_SAMPLER_CUBE = uguugl.GL_SAMPLER_CUBE
GL_DELETE_STATUS = uguugl.GL_DELETE_STATUS
GL_COMPILE_STATUS = uguugl.GL_COMPILE_STATUS
GL_LINK_STATUS = uguugl.GL_LINK_STATUS
GL_VALIDATE_STATUS = uguugl.GL_VALIDATE_STATUS
GL_INFO_LOG_LENGTH = uguugl.GL_INFO_LOG_LENGTH
GL_ATTACHED_SHADERS = uguugl.GL_ATTACHED_SHADERS
GL_ACTIVE_UNIFORMS = uguugl.GL_ACTIVE_UNIFORMS
GL_ACTIVE_UNIFORM_MAX_LENGTH = uguugl.GL_ACTIVE_UNIFORM_MAX_LENGTH
GL_SHADER_SOURCE_LENGTH = uguugl.GL_SHADER_SOURCE_LENGTH
GL_ACTIVE_ATTRIBUTES = uguugl.GL_ACTIVE_ATTRIBUTES
GL_ACTIVE_ATTRIBUTE_MAX_LENGTH = uguugl.GL_ACTIVE_ATTRIBUTE_MAX_LENGTH
GL_SHADING_LANGUAGE_VERSION = uguugl.GL_SHADING_LANGUAGE_VERSION
GL_CURRENT_PROGRAM = uguugl.GL_CURRENT_PROGRAM
GL_STENCIL_BACK_REF = uguugl.GL_STENCIL_BACK_REF
GL_STENCIL_BACK_VALUE_MASK = uguugl.GL_STENCIL_BACK_VALUE_MASK
GL_STENCIL_BACK_WRITEMASK = uguugl.GL_STENCIL_BACK_WRITEMASK
GL_FRAMEBUFFER_BINDING = uguugl.GL_FRAMEBUFFER_BINDING
GL_RENDERBUFFER_BINDING = uguugl.GL_RENDERBUFFER_BINDING
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE = uguugl.GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME = uguugl.GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL = uguugl.GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE = uguugl.GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE
GL_FRAMEBUFFER_COMPLETE = uguugl.GL_FRAMEBUFFER_COMPLETE
GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT = uguugl.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT
GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT = uguugl.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT
GL_FRAMEBUFFER_UNSUPPORTED = uguugl.GL_FRAMEBUFFER_UNSUPPORTED
GL_COLOR_ATTACHMENT0 = uguugl.GL_COLOR_ATTACHMENT0
GL_DEPTH_ATTACHMENT = uguugl.GL_DEPTH_ATTACHMENT
GL_STENCIL_ATTACHMENT = uguugl.GL_STENCIL_ATTACHMENT
GL_FRAMEBUFFER = uguugl.GL_FRAMEBUFFER
GL_RENDERBUFFER = uguugl.GL_RENDERBUFFER
GL_RENDERBUFFER_WIDTH = uguugl.GL_RENDERBUFFER_WIDTH
GL_RENDERBUFFER_HEIGHT = uguugl.GL_RENDERBUFFER_HEIGHT
GL_RENDERBUFFER_INTERNAL_FORMAT = uguugl.GL_RENDERBUFFER_INTERNAL_FORMAT
GL_STENCIL_INDEX8 = uguugl.GL_STENCIL_INDEX8
GL_RENDERBUFFER_RED_SIZE = uguugl.GL_RENDERBUFFER_RED_SIZE
GL_RENDERBUFFER_GREEN_SIZE = uguugl.GL_RENDERBUFFER_GREEN_SIZE
GL_RENDERBUFFER_BLUE_SIZE = uguugl.GL_RENDERBUFFER_BLUE_SIZE
GL_RENDERBUFFER_ALPHA_SIZE = uguugl.GL_RENDERBUFFER_ALPHA_SIZE
GL_RENDERBUFFER_DEPTH_SIZE = uguugl.GL_RENDERBUFFER_DEPTH_SIZE
GL_RENDERBUFFER_STENCIL_SIZE = uguugl.GL_RENDERBUFFER_STENCIL_SIZE



#cython: profile=False
# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import print_function

from gl1 cimport *
from gldraw cimport *

NONE = 0
BLIT = 1
BLEND = 2
IMAGEBLEND = 3

cdef void gl_clip(GLenum plane, GLdouble a, GLdouble b, GLdouble c, GLdouble d):
    """
    Utility function that takes care of setting up an OpenGL clip plane.
    """

    cdef GLdouble equation[4]

    equation[0] = a
    equation[1] = b
    equation[2] = c
    equation[3] = d
    glEnable(plane)
    glClipPlane(plane, equation)

cdef class FixedFunctionEnviron(Environ):
    """
    This is an OpenGL environment that uses the fixed-function pipeline.

    It requires ARB_texture_env_combine and ARB_texture_env_crossbar to
    work.
    """

    cdef object last
    cdef int last_ramp
    cdef int last_ramplen
    cdef object ramp_setup

    def init(self):

        # The last blend environ used.
        self.last = NONE

        # The last ramp length asked for.
        self.last_ramp = -1

        # The last ramplen actually used.
        self.last_ramplen = -1

        # A table that maps ramp lengths to the setup of the fixed-function
        # units.
        self.ramp_setup = [
            # Fields are:
            # ramp length,
            # unit 1 scale,
            # unit 2 function,
            # unit 2 scale,

            (256, 1.0, GL_REPLACE, 1.0),
            (128, 1.0, GL_REPLACE, 2.0),
            (64, 1.0, GL_REPLACE, 4.0),
            (32, 1.0, GL_ADD, 4.0),
            (16, 2.0, GL_ADD, 4.0),
            (8, 4.0, GL_ADD, 4.0),
            ]


    def deinit(self):
        """
        Called before changing the GL context.
        """

        return

    def disable(self, unit):
        """
        Disables the given texture combiner.
        """

        glActiveTextureARB(unit)
        glDisable(GL_TEXTURE_2D)

    # As this takes keyword arguments, it can't be a cdef function.
    def combine_mode(self, unit,
                     color_function=GL_MODULATE,
                     color_arg0=GL_TEXTURE,
                     color_arg1=GL_PREVIOUS_ARB,
                     color_arg2=GL_CONSTANT_ARB,
                     color_source0=GL_SRC_COLOR,
                     color_source1=GL_SRC_COLOR,
                     color_source2=GL_SRC_COLOR,
                     color_scale=1.0,
                     alpha_function=GL_MODULATE,
                     alpha_arg0=GL_TEXTURE,
                     alpha_arg1=GL_PREVIOUS_ARB,
                     alpha_arg2=GL_CONSTANT_ARB,
                     alpha_source0=GL_SRC_ALPHA,
                     alpha_source1=GL_SRC_ALPHA,
                     alpha_source2=GL_SRC_ALPHA,
                     alpha_scale=1.0,
                     enable=True):


        glActiveTextureARB(unit)

        if enable:
            glEnable(GL_TEXTURE_2D)
        else:
            glDisable(GL_TEXTURE_2D)

        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB)
        glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, color_function)
        glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, alpha_function)

        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, color_arg0)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, color_source0)
        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, color_arg1)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, color_source1)
        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, color_arg2)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, color_source2)

        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, alpha_arg0)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, alpha_source0)
        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, alpha_arg1)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, alpha_source1)
        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_ALPHA_ARB, alpha_arg2)
        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA_ARB, alpha_source2)

        glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, color_scale)
        glTexEnvf(GL_TEXTURE_ENV, GL_ALPHA_SCALE, alpha_scale)


    cdef void blit(self):

        if self.last != BLIT:

            # Set unit 0 to modulate.

            self.combine_mode(GL_TEXTURE0_ARB,
                              color_function=GL_MODULATE,
                              alpha_function=GL_MODULATE)


            # Disable units 1, 2 and 3.
            self.disable(GL_TEXTURE1_ARB)
            self.disable(GL_TEXTURE2_ARB)
            self.disable(GL_TEXTURE3_ARB)

            self.last = BLIT

    cdef void blend(self, double fraction):

        if self.last != BLEND:

            # Get texture 0.
            self.combine_mode(GL_TEXTURE0_ARB,
                              color_function=GL_REPLACE,
                              alpha_function=GL_REPLACE)

            # Use interpolate to combine texture 0 with texture 1, as
            # controlled by the constant of texture unit 1.
            self.combine_mode(GL_TEXTURE1_ARB,
                              color_function=GL_INTERPOLATE_ARB,
                              color_arg0=GL_TEXTURE1_ARB,
                              color_arg1=GL_TEXTURE0_ARB,
                              alpha_function=GL_INTERPOLATE_ARB,
                              alpha_arg0=GL_TEXTURE1_ARB,
                              alpha_arg1=GL_TEXTURE0_ARB)

            # Combine the interpolated result with the primary color, to
            # allow for tinting and alpha adjustments.
            self.combine_mode(GL_TEXTURE2_ARB,
                              color_function=GL_MODULATE,
                              color_arg0=GL_PREVIOUS_ARB,
                              color_arg1=GL_PRIMARY_COLOR_ARB,
                              alpha_function=GL_MODULATE,
                              alpha_arg0=GL_PREVIOUS_ARB,
                              alpha_arg1=GL_PRIMARY_COLOR_ARB,
                              enable=False)

            # Disable texture unit 3.
            self.disable(GL_TEXTURE3_ARB)

            self.last = BLEND

        cdef float *fractions = [ fraction, fraction, fraction, fraction ]

        glActiveTextureARB(GL_TEXTURE1_ARB)
        glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, fractions)



    cdef void imageblend(self, double fraction, int ramp):

        if self.last != IMAGEBLEND or self.last_ramp != ramp:

            # Figure out the details of the ramp.
            for i in self.ramp_setup:
                ramplen, t0scale, t1function, t1scale = i

                # We round the ramplen down, unless we're below the
                # smallest size, in which case we increase it.
                if ramplen <= ramp:
                    break

            # Unit 0 loads in the control image, and adds/subtracts an
            # offset to/from it. (We'll set it up to subtract here,
            # and change this later on as necessary.)
            #
            # It multiplies this result by up to 4.
            self.combine_mode(GL_TEXTURE0_ARB,
                              color_function=GL_SUBTRACT_ARB,
                              alpha_function=GL_SUBTRACT_ARB,
                              color_arg1=GL_CONSTANT_ARB,
                              alpha_arg1=GL_CONSTANT_ARB,
                              color_scale=t0scale,
                              alpha_scale=t0scale)

            # Unit 1 will multiply the result of stage 1 by up to 8.
            #
            # It also takes the first source texture, but we don't
            # access it directly yet.
            self.combine_mode(GL_TEXTURE1_ARB,
                              color_function=t1function,
                              alpha_function=t1function,
                              color_arg0=GL_PREVIOUS_ARB,
                              color_arg1=GL_PREVIOUS_ARB,
                              alpha_arg0=GL_PREVIOUS_ARB,
                              alpha_arg1=GL_PREVIOUS_ARB,
                              color_scale=t1scale,
                              alpha_scale=t1scale)

            # Unit 2 uses the result of unit 1 to interpolate between the
            # unit 1 and unit 2 textures.
            self.combine_mode(GL_TEXTURE2_ARB,
                              color_function=GL_INTERPOLATE_ARB,
                              alpha_function=GL_INTERPOLATE_ARB,
                              color_arg0=GL_TEXTURE2_ARB,
                              color_arg1=GL_TEXTURE1_ARB,
                              color_arg2=GL_PREVIOUS_ARB,
                              alpha_arg0=GL_TEXTURE2_ARB,
                              alpha_arg1=GL_TEXTURE1_ARB,
                              alpha_arg2=GL_PREVIOUS_ARB)

            # Finally, Unit 3 modulates the result of unit 2 with the color.
            self.combine_mode(GL_TEXTURE3_ARB,
                              color_function=GL_MODULATE,
                              color_arg0=GL_PREVIOUS_ARB,
                              color_arg1=GL_PRIMARY_COLOR_ARB,
                              alpha_function=GL_MODULATE,
                              alpha_arg0=GL_PREVIOUS_ARB,
                              alpha_arg1=GL_PRIMARY_COLOR_ARB,
                              enable=False)

            self.last = IMAGEBLEND
            self.last_ramp = ramp
            self.last_ramplen = ramplen

        # Compute the offset to apply to the alpha.
        start = -1.0
        end = self.last_ramplen / 256.0
        offset = start + ( end - start) * fraction

        # Decide if we're adding or subtracting.
        if offset < 0:
            function = GL_SUBTRACT_ARB
            offset = -offset
        else:
            function = GL_ADD

        # Setup unit 0 as appropriate.
        glActiveTextureARB(GL_TEXTURE0_ARB)
        glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, function)
        glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, function)

        cdef float *offsets = [ offset, offset, offset, offset ]

        glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, offsets)

    cdef void set_vertex(self, float *vertices):
        glVertexPointer(2, GL_FLOAT, 0, <GLubyte *> vertices)
        glEnableClientState(GL_VERTEX_ARRAY)

    cdef void set_texture(self, int unit, float *coords):

        if unit == 0:
            glClientActiveTextureARB(GL_TEXTURE0)
        elif unit == 1:
            glClientActiveTextureARB(GL_TEXTURE1)
        elif RENPY_THIRD_TEXTURE and unit == 2:
            glClientActiveTextureARB(GL_TEXTURE2)
        else:
            return

        if coords is not NULL:
            glTexCoordPointer(2, GL_FLOAT, 0, <GLubyte *> coords)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY)
        else:
            glDisableClientState(GL_TEXTURE_COORD_ARRAY)

    cdef void set_color(self, double r, double g, double b, double a):
        glColor4f(r, g, b, a)

    cdef void ortho(self, double left, double right, double bottom, double top, double near, double far):

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(left, right, bottom, top, near, far)
        glMatrixMode(GL_MODELVIEW)

    cdef void set_clip(self, tuple clip_box, GLDraw draw):

        cdef double minx, miny, maxx, maxy

        minx, miny, maxx, maxy = clip_box

        gl_clip(GL_CLIP_PLANE0, 1.0, 0.0, 0.0, -minx - .01)
        gl_clip(GL_CLIP_PLANE1, 0.0, 1.0, 0.0, -miny - .01)
        gl_clip(GL_CLIP_PLANE2, -1.0, 0.0, 0.0, maxx + .01)
        gl_clip(GL_CLIP_PLANE3, 0.0, -1.0, 0.0, maxy + .01)

    cdef void unset_clip(self, GLDraw draw):
        glDisable(GL_CLIP_PLANE0)
        glDisable(GL_CLIP_PLANE1)
        glDisable(GL_CLIP_PLANE2)
        glDisable(GL_CLIP_PLANE3)

    cdef void viewport(self, int x, int y, int width, int height):
        glViewport(x, y, width, height)
